# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009-2019 Dominik Kriegner <dominik.kriegner@gmail.com>
# Copyright (C) 2012 Tanja Etzelstorfer <tanja.etzelstorfer@jku.at>

"""
Classes decribing materials. Materials are devided with respect to their
crystalline state in either Amorphous or Crystal types.  While for most
materials their crystalline state is defined few materials are also included as
amorphous which can be useful for calculation of their optical properties.
"""
import abc
import copy
import numbers
import operator
import re
import warnings
from math import ceil, copysign

import numpy
import scipy.optimize
from pkg_resources import parse_version

from .. import config, math, utilities
from ..exception import InputError
from ..math import VecCross, VecDot, VecNorm
from . import cif, elements
from .atom import Atom
from .spacegrouplattice import WyckoffBase

numpy.seterr(divide='ignore', invalid='ignore')

map_ijkl2ij = {"00": 0, "11": 1, "22": 2,
               "12": 3, "20": 4, "01": 5,
               "21": 6, "02": 7, "10": 8}
map_ij2ijkl = {"0": [0, 0], "1": [1, 1], "2": [2, 2],
               "3": [1, 2], "4": [2, 0], "5": [0, 1],
               "6": [2, 1], "7": [0, 2], "8": [1, 0]}


def index_map_ijkl2ij(i, j):
    return map_ijkl2ij["%i%i" % (i, j)]


def index_map_ij2ijkl(ij):
    return map_ij2ijkl["%i" % ij]


def Cij2Cijkl(cij):
    """
    Converts the elastic constants matrix (tensor of rank 2) to
    the full rank 4 cijkl tensor.

    Parameters
    ----------
    cij :   array-like
        (6, 6) cij matrix

    Returns
    -------
    cijkl   ndarray
        (3, 3, 3, 3) cijkl tensor as numpy array
    """

    # first have to build a 9x9 matrix from the 6x6 one
    m = numpy.zeros((9, 9), dtype=numpy.double)
    m[0:6, 0:6] = cij[:, :]
    m[6:9, 0:6] = cij[3:6, :]
    m[0:6, 6:9] = cij[:, 3:6]
    m[6:9, 6:9] = cij[3:6, 3:6]

    # now create the full tensor
    cijkl = numpy.empty((3, 3, 3, 3), dtype=numpy.double)

    for i in range(0, 3):
        for j in range(0, 3):
            for k in range(0, 3):
                for l in range(0, 3):
                    mi = index_map_ijkl2ij(i, j)
                    mj = index_map_ijkl2ij(k, l)
                    cijkl[i, j, k, l] = m[mi, mj]
    return cijkl


def Cijkl2Cij(cijkl):
    """
    Converts the full rank 4 tensor of the elastic constants to
    the (6, 6) matrix of elastic constants.

    Parameters
    ----------
    cijkl   ndarray
        (3, 3, 3, 3) cijkl tensor as numpy array

    Returns
    -------
    cij :   array-like
        (6, 6) cij matrix
    """

    cij = numpy.empty((6, 6), dtype=numpy.double)

    for i in range(6):
        for j in range(6):
            ij = index_map_ij2ijkl(i)
            kl = index_map_ij2ijkl(j)
            cij[i, j] = cijkl[ij[0], ij[1], kl[0], kl[1]]

    return cij


class Material(utilities.ABC):
    """
    base class for all Materials. common properties of amorphous and
    crystalline materials are described by this class from which Amorphous and
    Crystal are derived from.
    """

    def __init__(self, name, cij=None):
        if cij is None:
            self.cij = numpy.zeros((6, 6), dtype=numpy.double)
            self.cijkl = numpy.zeros((3, 3, 3, 3), dtype=numpy.double)
        elif isinstance(cij, (tuple, list, numpy.ndarray)):
            self.cij = numpy.asarray(cij, dtype=numpy.double)
            self.cijkl = Cij2Cijkl(self.cij)
        else:
            raise TypeError("Elastic constants must be a list or numpy array!")

        self.name = name
        self.transform = None
        self._density = None

    def __getattr__(self, name):
        if name.startswith("c"):
            index = name[1:]
            if len(index) > 2:
                raise AttributeError("Cij indices must be between 1 and 6")

            i = int(index[0])
            j = int(index[1])

            if i > 6 or i < 1 or j > 6 or j < 1:
                raise AttributeError("Cij indices must be between 1 and 6")

            if callable(self.transform):
                cij = Cijkl2Cij(self.transform(Cij2Cijkl(self.cij)))
            else:
                cij = self.cij

            return cij[i - 1, j - 1]
        else:
            object.__getattribute__(self, name)

    def _getmu(self):
        return self.cij[3, 3]

    def _getlam(self):
        return self.cij[0, 1]

    def _getnu(self):
        return self.lam / 2. / (self.mu + self.lam)

    def _getdensity(self):
        return self._density

    density = property(_getdensity)
    mu = property(_getmu)
    lam = property(_getlam)
    nu = property(_getnu)

    @abc.abstractmethod
    def delta(self, en='config'):
        """
        abstract method which every implementation of a Material has to
        override
        """
        pass

    @abc.abstractmethod
    def ibeta(self, en='config'):
        """
        abstract method which every implementation of a Material has to
        override
        """
        pass

    def chi0(self, en='config'):
        """
        calculates the complex chi_0 values often needed in simulations.
        They are closely related to delta and beta
        (n = 1 + chi_r0/2 + i*chi_i0/2   vs.  n = 1 - delta + i*beta)
        """
        return (-2 * self.delta(en) + 2j * self.ibeta(en))

    def idx_refraction(self, en="config"):
        """
        function to calculate the complex index of refraction of a material
        in the x-ray range

        Parameters
        ----------
        en :    energy of the x-rays, if omitted the value from the
                xrayutilities configuration is used

        Returns
        -------
         n (complex)
        """
        n = 1. - self.delta(en) + 1.j * self.ibeta(en)
        return n

    def critical_angle(self, en='config', deg=True):
        """
        calculate critical angle for total external reflection

        Parameters
        ----------
        en :    float or str, optional
            energy of the x-rays in eV, if omitted the value from the
            xrayutilities configuration is used
        deg :   bool, optional
            return angle in degree if True otherwise radians (default:True)

        Returns
        -------
        float
            Angle of total external reflection
        """
        rn = 1. - self.delta(en)

        alphac = numpy.arccos(rn)
        if deg:
            alphac = numpy.degrees(alphac)

        return alphac

    def absorption_length(self, en='config'):
        """
        wavelength dependent x-ray absorption length defined as
        mu = lambda/(2*pi*2*beta) with lambda and beta as the x-ray
        wavelength and complex part of the refractive index respectively.

        Parameters
        ----------
        en :    float or str, optional
            energy of the x-rays in eV

        Returns
        -------
        float
            the absorption length in um
        """
        if isinstance(en, str) and en == 'config':
            en = utilities.energy(config.ENERGY)
        return utilities.en2lam(en) / (2 * numpy.pi * self.ibeta(en) * 2) / 1e4

    def __str__(self):
        ostr = "%s: %s\n" % (self.__class__.__name__, self.name)
        if numpy.any(self.cij):
            ostr += "Elastic tensor (6x6):\n"
            d = numpy.get_printoptions()
            numpy.set_printoptions(precision=2, linewidth=78, suppress=False)
            ostr += str(self.cij) + '\n'
            numpy.set_printoptions(d)

        return ostr


class Amorphous(Material):
    """
    amorphous materials are described by this class
    """

    def __init__(self, name, density, atoms=None, cij=None):
        """
        constructor of an amorphous material. The amorphous material is
        described by its density and atom composition.

        Parameters
        ----------
        name :      str
            name of the material. To allow automatic parsing of the chemical
            elements use the abbreviation of the chemical element from the
            periodic table. To specify alloys, use e.g. 'Ir0.2Mn0.8' or 'H2O'.
        density :   float
            mass density in kg/m^3
        atoms :     list, optional
            list of atoms together with their fractional content.  When the
            name is a simply chemical formula then this can be None.  To
            specify more complicated materials use [('Ir', 0.2), ('Mn', 0.8),
            ...].  Instead of the elements as string you can also use an Atom
            object.  If the contents to not add up to 1 they will be corrected
            without notice.
        cij :       array-like, optional
            elasticity matrix
        """
        super().__init__(name, cij)
        self._density = density
        self.base = list()
        if atoms is None:
            comp = Amorphous.parseChemForm(name)
            if config.VERBOSITY >= config.DEBUG:
                print("XU.materials.Amorphous: using '%s' as chemical formula"
                      % ''.join(['%s%.2f ' % (e.name, c) for e, c in comp]))
            for (e, c) in comp:
                self.base.append((e, c))
        else:
            frsum = numpy.sum([at[1] for at in atoms])
            for at, fr in atoms:
                if not isinstance(at, Atom):
                    a = getattr(elements, at)
                else:
                    a = at
                self.base.append((a, fr/frsum))

    @staticmethod
    def parseChemForm(cstring):
        """
        Parse a string containing a simple chemical formula and transform it to
        a list of elements together with their relative atomic fraction. e.g.
        'H2O' -> [(H, 2/3), (O, 1/3)], where H and O are the Element objects of
        Hydrogen and Oxygen. Note that every chemical element needs to start
        with a capital letter! Complicated formulas containing bracket are not
        supported!

        Parameters
        ----------
        cstring :   str
            string containing the chemical fomula

        Returns
        -------
        list of tuples
            chemical element and atomic fraction
        """
        if re.findall(r'[\(\)]', cstring):
            raise ValueError('unsupported chemical formula (%s) given.'
                             % cstring)
        elems = re.findall('[A-Z][^A-Z]*', cstring)
        r = re.compile(r"([a-zA-Z]+)([0-9\.]+)")
        ret = []
        csum = 0
        for e in elems:
            if r.match(e):
                elstr, cont = r.match(e).groups()
                cont = float(cont)
            else:
                elstr, cont = (e, 1.0)
            ret.append((elstr, cont))
            csum += cont
        for i, r in enumerate(ret):
            ret[i] = (getattr(elements, r[0]), r[1]/csum)
        return ret

    def _get_f(self, q, en):
        """
        optimized method to calculate the atomic scattering factor for all
        atoms in the unit cell by calling the database only as much as needed.

        Parameters
        ----------
        q :     float or array-like
            momentum transfer for which the atomic scattering factor should be
            calculated
        en :    float or str
            x-ray energy (eV)

        Returns
        -------
        list
            atomic scattering factors for every atom in the unit cell
        """
        f = {}
        for at, occ in self.base:
            if at.num not in f:
                f[at.num] = at.f(q, en)
        return [f[a.num] for a, o in self.base]

    def delta(self, en='config'):
        """
        function to calculate the real part of the deviation of the
        refractive index from 1 (n=1-delta+i*beta)

        Parameters
        ----------
        en :    float, array-like or str, optional
            energy of the x-rays in eV

        Returns
        -------
        float or array-like
        """
        re = scipy.constants.physical_constants['classical electron radius'][0]
        re *= 1e10
        if isinstance(en, str) and en == 'config':
            en = utilities.energy(config.ENERGY)

        lam = utilities.en2lam(en)
        delta = 0.
        m = 0.
        f = self._get_f(0., en)
        for (at, occ), fa in zip(self.base, f):
            delta += numpy.real(fa) * occ
            m += at.weight * occ

        delta *= re / (2 * numpy.pi) * lam ** 2 / (m / self.density) * 1e-30
        return delta

    def ibeta(self, en='config'):
        """
        function to calculate the imaginary part of the deviation
        of the refractive index from 1 (n=1-delta+i*beta)

        Parameters
        ----------
        en :    float, array-like or str, optional
            energy of the x-rays in eV

        Returns
        -------
        float or array-like
        """
        re = scipy.constants.physical_constants['classical electron radius'][0]
        re *= 1e10
        if isinstance(en, str) and en == 'config':
            en = utilities.energy(config.ENERGY)

        lam = utilities.en2lam(en)
        beta = 0.
        m = 0.
        f = self._get_f(0., en)
        for (at, occ), fa in zip(self.base, f):
            beta += numpy.imag(fa) * occ
            m += at.weight * occ

        beta *= re / (2 * numpy.pi) * lam ** 2 / (m / self.density) * 1e-30
        return beta

    def chi0(self, en='config'):
        """
        calculates the complex chi_0 values often needed in simulations.
        They are closely related to delta and beta
        (n = 1 + chi_r0/2 + i*chi_i0/2   vs.  n = 1 - delta + i*beta)
        """
        re = scipy.constants.physical_constants['classical electron radius'][0]
        re *= 1e10
        if isinstance(en, str) and en == 'config':
            en = utilities.energy(config.ENERGY)

        lam = utilities.en2lam(en)
        beta = 0.
        delta = 0.
        m = 0.
        f = self._get_f(0., en)
        for (at, occ), f0 in zip(self.base, f):
            beta += numpy.imag(f0) * occ
            delta += numpy.real(f0) * occ
            m += at.weight * occ

        beta *= re / (2 * numpy.pi) * lam ** 2 / (m / self.density) * 1e-30
        delta *= re / (2 * numpy.pi) * lam ** 2 / (m / self.density) * 1e-30
        return (-2 * delta + 2j * beta)

    def __str__(self):
        ostr = super().__str__()
        ostr += "density: %.2f\n" % self.density
        if self.base:
            ostr += "atoms: "
            for at, o in self.base:
                ostr += "(%s, %.3f) " % (at.name, o)
            ostr += "\n"

        return ostr


class Crystal(Material):
    """
    Crystalline materials are described by this class
    """

    def __init__(self, name, lat, cij=None, thetaDebye=None):
        super().__init__(name, cij)

        self.lattice = lat
        if isinstance(thetaDebye, numbers.Number):
            self.thetaDebye = float(thetaDebye)
        else:
            self.thetaDebye = thetaDebye

    @classmethod
    def fromCIF(cls, ciffilestr):
        """
        Create a Crystal from a CIF file. The default data-set from the cif
        file will be used to create the Crystal.

        Parameters
        ----------
        ciffilestr :    str, bytes
            filename of the CIF file or string representation of the CIF file

        Returns
        -------
        Crystal
        """
        cf = cif.CIFFile(ciffilestr)
        lat = cf.SGLattice()
        return cls(cf.data[cf._default_dataset].name, lat)

    def loadLatticefromCIF(self, ciffilestr):
        """
        load the unit cell data (lattice) from the CIF file. Other material
        properties stay unchanged.

        Parameters
        ----------
        ciffilestr :    str, bytes
            filename of the CIF file or string representation of the CIF file
        """
        cf = cif.CIFFile(ciffilestr)
        self.lattice = cf.SGLattice()

    def toCIF(self, ciffilename):
        """
        Export the Crystal to a CIF file.

        Parameters
        ----------
        ciffilename :  str
            filename of the CIF file
        """
        cif.cifexport(ciffilename, self)

    @property
    def a(self):
        return self.lattice.a

    @property
    def b(self):
        return self.lattice.b

    @property
    def c(self):
        return self.lattice.c

    @property
    def alpha(self):
        return self.lattice.alpha

    @property
    def beta(self):
        return self.lattice.beta

    @property
    def gamma(self):
        return self.lattice.gamma

    @property
    def a1(self):
        return self.lattice._ai[0, :]

    @property
    def a2(self):
        return self.lattice._ai[1, :]

    @property
    def a3(self):
        return self.lattice._ai[2, :]

    @property
    def B(self):
        return self.lattice.qtransform.matrix

    def __eq__(self, other):
        """
        compare if another Crystal instance is equal to the current one.
        Currently this considers only the lattice to be equal. Additional
        parameters like thetaDebye and the eleastic parameters are ignored.

        Parameters
        ----------
        other:  Crystal
            another instance of Crystal to compare
        """
        return self.lattice == other.lattice

    def Q(self, *hkl):
        """
        Return the Q-space position for a certain material.

        Parameters
        ----------
        hkl :   list or array-like
            Miller indices (or Q(h, k, l) is also possible)

        """
        return self.lattice.GetQ(*hkl)

    def HKL(self, *q):
        """
        Return the HKL-coordinates for a certain Q-space position.

        Parameters
        ----------
        q :   list or array-like
             Q-position. its also possible to use HKL(qx, qy, qz).

        """
        return self.lattice.GetHKL(*q)

    def chemical_composition(self, natoms=None, with_spaces=False, ndigits=2):
        """
        determine chemical composition from occupancy of atomic positions.

        Parameters
        ----------
        mat :       Crystal
            instance of Crystal
        natoms :    int, optional
            number of atoms to normalize the formula, if None some automatic
            normalization is attempted using the greatest common divisor of the
            number of atoms per unit cell. If the number of atoms of any
            element is fractional natoms=1 is used.
        with_spaces : bool, optional
            add spaces between the different entries in the output string for
            CIF combatibility
        ndigits :   int, optional
            number of digits to which floating point numbers are rounded to

        Returns
        -------
        str
            representation of the chemical composition
        """
        elem = {}
        for a in self.lattice.base():
            e = a[0].name
            occ = a[2]
            if e in elem:
                elem[e] += occ
            else:
                elem[e] = occ
        natom = sum([elem[e] for e in elem])
        isint = True
        for e in elem:
            if not float(elem[e]).is_integer():
                isint = False
        # determine number of atoms
        if not natoms:
            if isint:
                gcd = math.gcd([int(elem[e]) for e in elem])
                natoms = natom/gcd
            else:
                natoms = 1

        # generate output strig
        cstr = ''
        fmtstr = '%d' if isint else '%%.%df' % ndigits
        for e in elem:
            n = elem[e] / float(natom) * natoms
            cstr += e
            if n != 1:
                cstr += fmtstr % n
            cstr += ' ' if with_spaces else ''
        return cstr.strip()

    def environment(self, *pos, **kwargs):
        """
        Returns a list of neighboring atoms for a given position within the the
        unit cell.

        Parameters
        ----------
        pos :       list or array-like
            fractional coordinate in the unit cell
        maxdist :   float
            maximum distance wanted in the list of neighbors (default: 7)

        Returns
        -------
        list of tuples
            (distance, atomType, multiplicity) giving distance sorted list of
            atoms
        """

        valid_kwargs = {'maxdist': 'maximum distance needed in the output'}
        utilities.check_kwargs(kwargs, valid_kwargs, 'Crystal.environment')
        maxdist = kwargs.get('maxdist', 7)

        if len(pos) < 3:
            pos = pos[0]
            if len(pos) < 3:
                raise InputError("need 3 coordinates of the "
                                 "reference position")

        refpos = self.a1 * \
            pos[0] + self.a2 * pos[1] + self.a3 * pos[2]

        lst = []
        Na = 2 * int(ceil(maxdist / math.VecNorm(self.a1)))
        Nb = 2 * int(ceil(maxdist / math.VecNorm(self.a2)))
        Nc = 2 * int(ceil(maxdist / math.VecNorm(self.a3)))
        if self.lattice.nsites > 0:
            for a, p, o, b in self.lattice.base():
                ucpos = (self.a1 * p[0] + self.a2 * p[1] + self.a3 * p[2])
                for i in range(-Na, Na + 1):
                    for j in range(-Nb, Nb + 1):
                        for k in range(-Nc, Nc + 1):
                            atpos = ucpos + (self.a1 * i + self.a2 * j +
                                             self.a3 * k)
                            distance = math.VecNorm(atpos - refpos)
                            if distance <= maxdist:
                                lst.append((distance, a, o))
        else:
            for i in range(-Na, Na + 1):
                for j in range(-Nb, Nb + 1):
                    for k in range(-Nc, Nc + 1):
                        atpos = (self.a1 * i + self.a2 * j + self.a3 * k)
                        distance = math.VecNorm(atpos - refpos)
                        if distance <= maxdist:
                            lst.append((distance, '__dummy__', 1.))

        # sort
        lst.sort(key=operator.itemgetter(1))
        lst.sort(key=operator.itemgetter(0))
        rl = []
        if len(lst) >= 1:
            mult = lst[0][2]
        else:
            return rl
        for i in range(1, len(lst)):
            if (abs(lst[i - 1][0] - lst[i][0]) < config.EPSILON and
                    lst[i - 1][1] == lst[i][1]):
                mult += lst[i - 1][2]  # add occupancy
            else:
                rl.append((lst[i - 1][0], lst[i - 1][1], mult))
                mult = lst[i][2]
        rl.append((lst[-1][0], lst[-1][1], mult))

        return rl

    def planeDistance(self, *hkl):
        """
        determines the lattice plane spacing for the planes specified by (hkl)

        Parameters
        ----------
        h, k, l :  list, tuple or floats
            Miller indices of the lattice planes given either as list, tuple or
            seperate arguments

        Returns
        -------
        float
            the lattice plane spacing

        Examples
        --------
        >>> xu.materials.Si.planeDistance(0, 0, 4)
        1.3577600000000001

        or

        >>> xu.materials.Si.planeDistance((1, 1, 1))
        3.1356124059796255
        """
        if len(hkl) < 3:
            hkl = hkl[0]
            if len(hkl) < 3:
                raise InputError("need 3 indices for the lattice point")

        return 2 * numpy.pi / math.VecNorm(self.Q(hkl))

    def _getdensity(self):
        """
        calculates the mass density of an material from the mass of the atoms
        in the unit cell.

        Returns
        -------
        float
            mass density in kg/m^3
        """
        m = 0.
        for at, pos, occ, b in self.lattice.base():
            m += at.weight * occ

        return m / self.lattice.UnitCellVolume() * 1e30

    density = property(_getdensity)

    def _get_f(self, q, en):
        """
        optimized method to calculate the atomic scattering factor for all
        atoms in the unit cell by calling the database only as much as needed.

        Parameters
        ----------
        q :     float or array-like
            momentum transfer for which the atomic scattering factor should be
            calculated
        en :    float or str
            x-ray energy (eV)

        Returns
        -------
        list
            atomic scattering factors for every atom in the unit cell
        """
        f = {}
        if self.lattice.nsites > 0:
            for at, pos, occ, b in self.lattice.base():
                if at.num not in f:
                    f[at.num] = at.f(q, en)
            return [f[a.num] for a, p, o, b in self.lattice.base()]
        else:
            return None

    def _get_lamen(self, en):
        if isinstance(en, str) and en == 'config':
            en = utilities.energy(config.ENERGY)
        lam = utilities.en2lam(en)
        return lam, en

    def delta(self, en='config'):
        """
        function to calculate the real part of the deviation of the
        refractive index from 1 (n=1-delta+i*beta)

        Parameters
        ----------
        en :    float or str, optional
            x-ray energy eV, if omitted the value from the xrayutilities
            configuration is used

        Returns
        -------
        float
        """
        re = scipy.constants.physical_constants['classical electron radius'][0]
        re *= 1e10

        lam, en = self._get_lamen(en)
        delta = 0.
        f = self._get_f(0, en)
        for (at, pos, occ, b), fa in zip(self.lattice.base(), f):
            delta += numpy.real(fa) * occ

        delta *= re / (2 * numpy.pi) * lam ** 2 / \
            self.lattice.UnitCellVolume()
        return delta

    def ibeta(self, en='config'):
        """
        function to calculate the imaginary part of the deviation
        of the refractive index from 1 (n=1-delta+i*beta)

        Parameters
        ----------
        en :    float or str, optional
            x-ray energy eV, if omitted the value from the xrayutilities
            configuration is used

        Returns
        -------
        float
        """
        re = scipy.constants.physical_constants['classical electron radius'][0]
        re *= 1e10

        lam, en = self._get_lamen(en)
        beta = 0.
        f = self._get_f(0, en)
        for (at, pos, occ, b), fa in zip(self.lattice.base(), f):
            beta += numpy.imag(fa) * occ

        beta *= re / (2 * numpy.pi) * lam ** 2 / self.lattice.UnitCellVolume()
        return beta

    def chi0(self, en='config'):
        """
        calculates the complex chi_0 values often needed in simulations.
        They are closely related to delta and beta
        (n = 1 + chi_r0/2 + i*chi_i0/2   vs.  n = 1 - delta + i*beta)
        """
        re = scipy.constants.physical_constants['classical electron radius'][0]
        re *= 1e10

        lam, en = self._get_lamen(en)
        beta = 0.
        delta = 0.
        if self.lattice.nsites > 0:
            f = self._get_f(0, en)
            for (at, pos, occ, b), f0 in zip(self.lattice.base(), f):
                beta += numpy.imag(f0) * occ
                delta += numpy.real(f0) * occ

        v = self.lattice.UnitCellVolume()
        beta *= re / (2 * numpy.pi) * lam ** 2 / v
        delta *= re / (2 * numpy.pi) * lam ** 2 / v
        return (-2 * delta + 2j * beta)

    def _debyewallerfactor(self, temp, qnorm):
        """
        Calculate the Debye Waller temperature factor according to the Debye
        temperature

        Parameters
        ----------
        temp :      float
            actual temperature (K)
        qnorm :     float or array-like
            norm of the q-vector(s) for which the factor should be calculated

        Returns
        -------
        float or array-like
            the Debye Waller factor(s) with the same shape as qnorm
        """
        if temp != 0 and self.thetaDebye:
            # W(q) = 3/2* hbar^2*q^2/(m*kB*tD) * (D1(tD/T)/(tD/T) + 1/4)
            # DWF = exp(-W(q)) consistent with Vaclav H. and several books
            hbar = scipy.constants.hbar
            kb = scipy.constants.Boltzmann
            x = self.thetaDebye / float(temp)
            m = 0.
            im = 0
            for a, p, o, b in self.lattice.base():
                m += a.weight
                im += 1
            m = m / float(im)
            exponentf = 3 / 2. * hbar ** 2 * 1.0e20 / \
                (m * kb * self.thetaDebye) * (math.Debye1(x) / x + 0.25)
            if config.VERBOSITY >= config.DEBUG:
                print("XU.materials.Crystal: DWF = exp(-W*q**2) W= %g"
                      % exponentf)
            dwf = numpy.exp(-exponentf * qnorm ** 2)
        else:
            dwf = 1.0
        return dwf

    def chih(self, q, en='config', temp=0, polarization='S'):
        """
        calculates the complex polarizability of a material for a certain
        momentum transfer and energy

        Parameters
        ----------
        q :     list, tuple or array-like
            momentum transfer vector in (1/A)
        en :    float or str, optional
            x-ray energy eV, if omitted the value from the xrayutilities
            configuration is used
        temp :  float, optional
            temperature used for Debye-Waller-factor calculation
        polarization :  {'S', 'P'}, optional
            sigma or pi polarization

        Returns
        -------
        tuple
            (abs(chih_real), abs(chih_imag)) complex polarizability
        """

        if isinstance(q, (list, tuple)):
            q = numpy.array(q, dtype=numpy.double)
        elif isinstance(q, numpy.ndarray):
            pass
        else:
            raise TypeError("q must be a list or numpy array!")
        qnorm = math.VecNorm(q)

        if isinstance(en, str) and en == 'config':
            en = utilities.energy(config.ENERGY)

        if polarization not in ('S', 'P'):
            raise ValueError("polarization must be 'S':sigma or 'P': pi!")

        if self.lattice.nsites == 0:
            return (0, 0)

        dwf = self._debyewallerfactor(temp, qnorm)

        sr = 0. + 0.j
        si = 0. + 0.j
        # a: atom, p: position, o: occupancy, b: temperature-factor
        f = self._get_f(qnorm, en)
        for (a, p, o, b), F in zip(self.lattice.base(), f):
            r = self.lattice.GetPoint(p)
            if temp == 0:
                dwf = numpy.exp(-b * qnorm ** 2 / (4 * numpy.pi) ** 2)
            fr = numpy.real(F) * o
            fi = numpy.imag(F) * o
            sr += fr * numpy.exp(-1.j * math.VecDot(q, r)) * dwf
            si += fi * numpy.exp(-1.j * math.VecDot(q, r)) * dwf

        # classical electron radius
        c = scipy.constants
        r_e = 1 / (4 * numpy.pi * c.epsilon_0) * c.e ** 2 / \
            (c.electron_mass * c.speed_of_light ** 2) * 1e10
        lam = utilities.en2lam(en)

        fact = -lam ** 2 * r_e / (numpy.pi * self.lattice.UnitCellVolume())
        rchi = numpy.abs(fact * sr)
        ichi = numpy.abs(fact * si)
        if polarization == 'P':
            theta = numpy.arcsin(qnorm * utilities.en2lam(en) / (4*numpy.pi))
            rchi *= numpy.cos(2 * theta)
            ichi *= numpy.cos(2 * theta)

        return rchi, ichi

    def dTheta(self, Q, en='config'):
        """
        function to calculate the refractive peak shift

        Parameters
        ----------
        Q :     list, tuple or array-like
            momentum transfer vector (1/A)
        en :    float or str, optional
            x-ray energy eV, if omitted the value from the xrayutilities
            configuration is used

        Returns
        -------
        float
            peak shift in degree
        """

        if isinstance(en, str) and en == 'config':
            en = utilities.energy(config.ENERGY)
        lam = utilities.en2lam(en)
        dth = numpy.degrees(
            2 * self.delta(en) / numpy.sin(2 * numpy.arcsin(
                lam * VecNorm(Q) / (4 * numpy.pi))))
        return dth

    def __str__(self):
        ostr = super().__str__()
        ostr += "Lattice:\n"
        ostr += str(self.lattice)
        return ostr

    def StructureFactor(self, q, en='config', temp=0):
        """
        calculates the structure factor of a material
        for a certain momentum transfer and energy
        at a certain temperature of the material

        Parameters
        ----------
        q :     list, tuple or array-like
            vectorial momentum transfer
        en :    float or str, optional
            x-ray energy eV, if omitted the value from the xrayutilities
            configuration is used
        temp :  float
            temperature used for Debye-Waller-factor calculation

        Returns
        -------
        complex
            the complex structure factor
        """

        if isinstance(q, (list, tuple)):
            q = numpy.array(q, dtype=numpy.double)
        elif isinstance(q, numpy.ndarray):
            pass
        else:
            raise TypeError("q must be a list or numpy array!")

        if isinstance(en, str) and en == 'config':
            en = utilities.energy(config.ENERGY)

        if self.lattice.nsites == 0:
            return 1.

        qnorm = math.VecNorm(q)
        dwf = self._debyewallerfactor(temp, qnorm)

        s = 0. + 0.j
        f = self._get_f(qnorm, en)
        # a: atom, p: position, o: occupancy, b: temperature-factor
        for (a, p, o, b), fq in zip(self.lattice.base(), f):
            r = self.lattice.GetPoint(p)
            if temp == 0:
                dwf = numpy.exp(-b * qnorm ** 2 /
                                (4 * numpy.pi) ** 2)
            s += fq * o * numpy.exp(-1.j * math.VecDot(q, r)) * dwf

        return s

    def StructureFactorForEnergy(self, q0, en, temp=0):
        """
        calculates the structure factor of a material
        for a certain momentum transfer and a bunch of energies

        Parameters
        ----------
        q0 :    list, tuple or array-like
            vectorial momentum transfer
        en :    list, tuple or array-like
            energy values in eV
        temp :  float
            temperature used for Debye-Waller-factor calculation

        Returns
        -------
        array-like
            complex valued structure factor array
        """
        if isinstance(q0, (list, tuple)):
            q = numpy.array(q0, dtype=numpy.double)
        elif isinstance(q0, numpy.ndarray):
            q = q0
        else:
            raise TypeError("q must be a list or numpy array!")
        qnorm = math.VecNorm(q)

        if isinstance(en, (list, tuple)):
            en = numpy.array(en, dtype=numpy.double)
        elif isinstance(en, numpy.ndarray):
            pass
        else:
            raise TypeError("Energy data must be provided as a list "
                            "or numpy array!")

        if self.lattice.nsites == 0:
            return numpy.ones(len(en))

        dwf = self._debyewallerfactor(temp, qnorm)

        s = 0. + 0.j
        f = self._get_f(qnorm, en)
        # a: atom, p: position, o: occupancy, b: temperature-factor
        for (a, p, o, b), fq in zip(self.lattice.base(), f):
            if temp == 0:
                dwf = numpy.exp(-b * qnorm ** 2 / (4 * numpy.pi) ** 2)
            r = self.lattice.GetPoint(p)
            s += fq * o * dwf * numpy.exp(-1.j * math.VecDot(q, r))

        return s

    def StructureFactorForQ(self, q, en0='config', temp=0):
        """
        calculates the structure factor of a material
        for a bunch of momentum transfers and a certain energy

        Parameters
        ----------
        q :     list of vectors or array-like
            vectorial momentum transfers; list of vectores (list, tuple or
            array) of length 3
            e.g.: (Si.Q(0, 0, 4), Si.Q(0, 0, 4.1),...) or
            numpy.array([Si.Q(0, 0, 4), Si.Q(0, 0, 4.1)])
        en0 :   float or str, optional
            x-ray energy eV, if omitted the value from the xrayutilities
            configuration is used
        temp :  float
            temperature used for Debye-Waller-factor calculation

        Returns
        -------
        array-like
            complex valued structure factor array
        """
        if isinstance(q, (list, tuple, numpy.ndarray)):
            q = numpy.asarray(q, dtype=numpy.double)
        else:
            raise TypeError("q must be a list or numpy array!")
        if len(q.shape) != 2:
            raise ValueError("q does not have the correct shape (shape = %s)"
                             % str(q.shape))
        qnorm = numpy.linalg.norm(q, axis=1)

        if isinstance(en0, str) and en0 == 'config':
            en0 = utilities.energy(config.ENERGY)

        if self.lattice.nsites == 0:
            return numpy.ones(len(q))

        dwf = self._debyewallerfactor(temp, qnorm)

        s = 0. + 0.j
        f = self._get_f(qnorm, en0)
        # a: atom, p: position, o: occupancy, b: temperature-factor
        for (a, p, o, b), fq in zip(self.lattice.base(), f):
            if temp == 0:
                dwf = numpy.exp(-b * qnorm ** 2 / (4 * numpy.pi) ** 2)

            r = self.lattice.GetPoint(p)
            s += fq * o * numpy.exp(-1.j * numpy.dot(q, r)) * dwf

        return s

    def ApplyStrain(self, strain):
        """
        Applies a certain strain on the lattice of the material. The result is
        a change in the base vectors of the real space as well as reciprocal
        space lattice.  The full strain matrix (3x3) needs to be given.

        Note:
            NO elastic response of the material will be considered!
        """
        # let strain act on the unit cell vectors
        self.lattice.ApplyStrain(strain)

    def GetMismatch(self, mat):
        """
        Calculate the mismatch strain between the material and a second
        material
        """
        raise NotImplementedError("XU.material.GetMismatch: "
                                  "not implemented yet")

    def distances(self):
        """
        function to obtain distances of atoms in the crystal up to the unit
        cell size (largest value of a, b, c is the cut-off)

        returns a list of tuples with distance d and number of occurence n
        [(d1, n1), (d2, n2),...]

        Note:
            if the base of the material is empty the list will be empty
        """

        if self.lattice.nsites == 0:
            return []

        cutoff = numpy.max((self.lattice.a, self.lattice.b, self.lattice.c))

        tmp_data = []

        for at1 in self.lattice.base():
            for at2 in self.lattice.base():
                dis = math.VecNorm(self.lattice.GetPoint(at1[1] - at2[1]))
                dis2 = math.VecNorm(self.lattice.GetPoint(
                    at1[1] - at2[1] + numpy.array((1, 0, 0))))
                dis3 = math.VecNorm(self.lattice.GetPoint(
                    at1[1] - at2[1] + numpy.array((0, 1, 0))))
                dis4 = math.VecNorm(self.lattice.GetPoint(
                    at1[1] - at2[1] + numpy.array((0, 0, 1))))
                dis5 = math.VecNorm(self.lattice.GetPoint(
                    at1[1] - at2[1] + numpy.array((-1, 0, 0))))
                dis6 = math.VecNorm(self.lattice.GetPoint(
                    at1[1] - at2[1] + numpy.array((0, -1, 0))))
                dis7 = math.VecNorm(self.lattice.GetPoint(
                    at1[1] - at2[1] + numpy.array((0, 0, -1))))
                distances = sorted([dis, dis2, dis3, dis4, dis5, dis6, dis7])

                for dis in distances:
                    if dis < cutoff:
                        tmp_data.append(dis)

        # sort the list and compress equal entries
        tmp_data.sort()

        self._distances = [0]
        self._dis_hist = [0]
        for dis in tmp_data:
            if numpy.round(dis - self._distances[-1],
                           int(abs(numpy.log10(config.EPSILON)))) == 0:
                self._dis_hist[-1] += 1
            else:
                self._distances.append(dis)
                self._dis_hist.append(1)

        # create return value
        ret = []
        for i in range(len(self._distances)):
            ret.append((self._distances[i], self._dis_hist[i]))

        return ret

    def show_unitcell(self, fig=None, subplot=111, scale=0.6, complexity=11,
                      linewidth=2):
        """
        primitive visualization of the unit cell using matplotlibs basic 3D
        functionality -> expect rendering inaccuracies!

        Note:
            For more precise visualization export to CIF and use a proper
            crystal structure viewer.

        Parameters
        ----------
        fig :   matplotlib Figure or None, optional
        subplot :   int or list, optional
            subplot to use for the visualization. This argument of fowarded to
            the first argument of matplotlibs `add_subplot` function
        scale :     float, optional
            scale the size of the atoms by this additional factor. By default
            the size of the atoms corresponds to 60% of their atomic radius.
        complexity :    int, optional
            number of steps to approximate the atoms as spheres. higher values
            cause significant slower plotting.
        linewidth :     float, optional
            line thickness of the unit cell outline
        """
        plot, plt = utilities.import_matplotlib_pyplot('XU.materials')
        try:
            import mpl_toolkits.mplot3d
        except ImportError:
            plot = False

        if not plot:
            print('matplotlib (including mplot3d) needed for show_unitcell()')
            return

        if fig is None:
            fig = plt.figure()
        ax = fig.add_subplot(subplot, projection='3d')

        phi, theta = numpy.mgrid[0:numpy.pi:1j*complexity,
                                 0:2*numpy.pi:1j*complexity]

        for a, pos, occ, b in self.lattice.base():
            r = a.radius * scale
            for i in range(-1, 2):
                for j in range(-1, 2):
                    for k in range(-1, 2):
                        atpos = (pos + [i, j, k])
                        inunitcell = True
                        for l in range(3):
                            if (atpos[l] < -config.EPSILON or
                                    atpos[l] > 1+config.EPSILON):
                                inunitcell = False
                        if inunitcell:
                            vecpos = atpos[0]*self.a1 + atpos[1]*self.a2 +\
                                     atpos[2]*self.a3
                            x = r*numpy.sin(phi)*numpy.cos(theta) + vecpos[0]
                            y = r*numpy.sin(phi)*numpy.sin(theta) + vecpos[1]
                            z = r*numpy.cos(phi) + vecpos[2]
                            ax.plot_surface(x, y, z,  rstride=1, cstride=1,
                                            color=a.color, alpha=occ,
                                            linewidth=0)

        # plot unit cell outlines
        ax.plot([0, self.a1[0]], [0, self.a1[1]], [0, self.a1[2]], color='k',
                lw=linewidth)
        ax.plot([0, self.a2[0]], [0, self.a2[1]], [0, self.a2[2]], color='k',
                lw=linewidth)
        ax.plot([0, self.a3[0]], [0, self.a3[1]], [0, self.a3[2]], color='k',
                lw=linewidth)
        ax.plot([self.a1[0], self.a1[0]+self.a2[0]],
                [self.a1[1], self.a1[1]+self.a2[1]],
                [self.a1[2], self.a1[2]+self.a2[2]], color='k', lw=linewidth)
        ax.plot([self.a1[0], self.a1[0]+self.a3[0]],
                [self.a1[1], self.a1[1]+self.a3[1]],
                [self.a1[2], self.a1[2]+self.a3[2]], color='k', lw=linewidth)
        ax.plot([self.a2[0], self.a1[0]+self.a2[0]],
                [self.a2[1], self.a1[1]+self.a2[1]],
                [self.a2[2], self.a1[2]+self.a2[2]], color='k', lw=linewidth)
        ax.plot([self.a2[0], self.a2[0]+self.a3[0]],
                [self.a2[1], self.a2[1]+self.a3[1]],
                [self.a2[2], self.a2[2]+self.a3[2]], color='k', lw=linewidth)
        ax.plot([self.a3[0], self.a1[0]+self.a3[0]],
                [self.a3[1], self.a1[1]+self.a3[1]],
                [self.a3[2], self.a1[2]+self.a3[2]], color='k', lw=linewidth)
        ax.plot([self.a3[0], self.a2[0]+self.a3[0]],
                [self.a3[1], self.a2[1]+self.a3[1]],
                [self.a3[2], self.a2[2]+self.a3[2]], color='k', lw=linewidth)
        ax.plot([self.a1[0]+self.a2[0], self.a1[0]+self.a2[0]+self.a3[0]],
                [self.a1[1]+self.a2[1], self.a1[1]+self.a2[1]+self.a3[1]],
                [self.a1[2]+self.a2[2], self.a1[2]+self.a2[2]+self.a3[2]],
                color='k', lw=linewidth)
        ax.plot([self.a1[0]+self.a3[0], self.a1[0]+self.a2[0]+self.a3[0]],
                [self.a1[1]+self.a3[1], self.a1[1]+self.a2[1]+self.a3[1]],
                [self.a1[2]+self.a3[2], self.a1[2]+self.a2[2]+self.a3[2]],
                color='k', lw=linewidth)
        ax.plot([self.a2[0]+self.a3[0], self.a1[0]+self.a2[0]+self.a3[0]],
                [self.a2[1]+self.a3[1], self.a1[1]+self.a2[1]+self.a3[1]],
                [self.a2[2]+self.a3[2], self.a1[2]+self.a2[2]+self.a3[2]],
                color='k', lw=linewidth)

        if parse_version(plt.matplotlib.__version__) < parse_version('3.1.0'):
            ax.set_aspect("equal")

        if config.VERBOSITY >= config.INFO_LOW:
            warnings.warn("show_unitcell: 3D projection might appear distorted"
                          "(limited 3D capabilities of matplotlib!). Use CIF "
                          "export and other viewers for better visualization.")
        plt.tight_layout()


def CubicElasticTensor(c11, c12, c44):
    """
    Assemble the 6x6 matrix of elastic constants for a cubic material from the
    three independent components of a cubic crystal

    Parameters
    ----------
    c11, c12, c44 : float
        independent components of the elastic tensor of cubic materials

    Returns
    -------
   cij :    ndarray
        6x6 matrix with elastic constants
    """
    m = numpy.zeros((6, 6), dtype=numpy.double)
    m[0, 0] = c11
    m[1, 1] = c11
    m[2, 2] = c11
    m[3, 3] = c44
    m[4, 4] = c44
    m[5, 5] = c44
    m[0, 1] = m[0, 2] = c12
    m[1, 0] = m[1, 2] = c12
    m[2, 0] = m[2, 1] = c12

    return m


def HexagonalElasticTensor(c11, c12, c13, c33, c44):
    """
    Assemble the 6x6 matrix of elastic constants for a hexagonal material from
    the five independent components of a hexagonal crystal

    Parameters
    ----------
    c11, c12, c13, c33, c44 :   float
        independent components of the elastic tensor of a hexagonal material

    Returns
    -------
   cij :    ndarray
        6x6 matrix with elastic constants
    """
    m = numpy.zeros((6, 6), dtype=numpy.double)
    m[0, 0] = m[1, 1] = c11
    m[2, 2] = c33
    m[3, 3] = m[4, 4] = c44
    m[5, 5] = 0.5 * (c11 - c12)
    m[0, 1] = m[1, 0] = c12
    m[0, 2] = m[1, 2] = m[2, 0] = m[2, 1] = c13

    return m


def WZTensorFromCub(c11ZB, c12ZB, c44ZB):
    """
    Determines the hexagonal elastic tensor from the values of the cubic
    elastic tensor under the assumptions presented in Phys. Rev. B 6, 4546
    (1972), which are valid for the WZ <-> ZB polymorphs.

    Parameters
    ----------
    c11, c12, c44 : float
        independent components of the elastic tensor of cubic materials

    Returns
    -------
    cij :   ndarray
        6x6 matrix with elastic constants

    Implementation according to a patch submitted by Julian Stangl
    """
    # matrix conversions: cubic (111) to hexagonal (001) direction
    P = (1 / 6.) * numpy.array([[3, 3, 6],
                                [2, 4, 8],
                                [1, 5, -2],
                                [2, 4, -4],
                                [2, -2, 2],
                                [1, -1, 4]])
    Q = (1 / (3 * numpy.sqrt(2))) * numpy.array([1, -1, -2])

    cZBvec = numpy.array([c11ZB, c12ZB, c44ZB])
    cWZvec_BAR = numpy.dot(P, cZBvec)
    delta = numpy.dot(Q, cZBvec)
    D = numpy.array([delta**2 / cWZvec_BAR[2], 0, -delta**2 / cWZvec_BAR[2],
                     0, delta**2 / cWZvec_BAR[0], delta**2 / cWZvec_BAR[2]])
    cWZvec = cWZvec_BAR - D.T

    return HexagonalElasticTensor(cWZvec[0], cWZvec[2], cWZvec[3],
                                  cWZvec[1], cWZvec[4])


class Alloy(Crystal):
    """
    alloys two materials from the same crystal system. If the materials have
    the same space group the Wyckoff positions within the unit cell will also
    reflect the alloying.
    """

    def __init__(self, matA, matB, x):
        self.check_compatibility(matA, matB)
        lat = copy.deepcopy(matA.lattice)
        super().__init__("None", lat, matA.cij)
        self.matA = matA
        self.matB = matB
        self._setxb(x)

    @staticmethod
    def check_compatibility(matA, matB):
        csA = matA.lattice.crystal_system.split(':')[0]
        csB = matB.lattice.crystal_system.split(':')[0]
        if csA != csB:
            raise InputError("Crystal systems of the two materials are "
                             "incompatible!")

    @staticmethod
    def lattice_const_AB(latA, latB, x, name=''):
        """
        method to calculated the interpolation of lattice parameters and unit
        cell angles of the Alloy. By default linear interpolation between the
        value of material A and B is performed.

        Parameters
        ----------
        latA, latB :    float or vector
            property (lattice parameter/angle) of material A and B.  A property
            can be a scalar or vector.
        x :             float
            fraction of material B in the alloy.
        name :          str, optional
            label of the property which is interpolated. Can be 'a', 'b', 'c',
            'alpha', 'beta', or 'gamma'.
        """
        return (latB - latA) * x + latA

    def _getxb(self):
        return self._xb

    def _setxb(self, x):
        self._xb = x
        self.name = ("%s(%2.2f)%s(%2.2f)"
                     % (self.matA.name, 1-x, self.matB.name, x))
        # modify the free parameters of the lattice
        for k in self.lattice.free_parameters:
            setattr(self.lattice, k,
                    self.lattice_const_AB(getattr(self.matA, k),
                                          getattr(self.matB, k), x, name=k))
        # set elastic constants
        self.cij = (self.matB.cij - self.matA.cij) * x + self.matA.cij
        self.cijkl = (self.matB.cijkl - self.matA.cijkl) * x + self.matA.cijkl
        # alloying in unit cell
        if self.matA.lattice.space_group == self.matB.lattice.space_group:
            self.lattice._wbase = WyckoffBase()
            for a, wp, o, b in self.matA.lattice._wbase:
                self.lattice._wbase.append(a, wp, occ=o*(1-x), b=b)
            for a, wp, o, b in self.matB.lattice._wbase:
                if (a, wp, o, b) in self.lattice._wbase:
                    idx = self.lattice._wbase.index((a, wp, o, b))
                    occ = self.lattice._wbase[idx][2]
                    self.lattice._wbase[idx] = (a, wp, occ+o*x, b)
                else:
                    self.lattice._wbase.append(a, wp, occ=o*x, b=b)

    x = property(_getxb, _setxb)

    def _checkfinitenumber(self, arg, name=""):
        if isinstance(arg, numbers.Number) and numpy.isfinite(arg):
            return float(arg)
        else:
            raise TypeError("argument (%s) must be a scalar!" % name)

    def _checkarray(self, arg, name=""):
        if isinstance(arg, (list, tuple, numpy.ndarray)):
            return numpy.asarray(arg, dtype=numpy.double)
        else:
            raise TypeError("argument (%s) must be of type "
                            "list, tuple or numpy.ndarray" % name)

    def _definehelpers(self, hkl, cijA, cijB):
        """
        define helper functions for solving the content from reciprocal space
        positions
        """
        def a1(x):
            return self.lattice_const_AB(self.matA.a1, self.matB.a1,
                                         x, name='a')

        def a2(x):
            return self.lattice_const_AB(self.matA.a2, self.matB.a2,
                                         x, name='b')

        def a3(x):
            return self.lattice_const_AB(self.matA.a3, self.matB.a3,
                                         x, name='c')

        def V(x):
            return numpy.dot(a3(x), numpy.cross(a1(x), a2(x)))

        def b1(x):
            return 2 * numpy.pi / V(x) * numpy.cross(a2(x), a3(x))

        def b2(x):
            return 2 * numpy.pi / V(x) * numpy.cross(a3(x), a1(x))

        def b3(x):
            return 2 * numpy.pi / V(x) * numpy.cross(a1(x), a2(x))

        def qhklx(x):
            return hkl[0] * b1(x) + hkl[1] * b2(x) + hkl[2] * b3(x)

        def frac(x):
            return ((cijB[0, 2] + cijB[1, 2] - (cijA[0, 2] + cijA[1, 2])) * x +
                    (cijA[0, 2] + cijA[1, 2])) / \
                   ((cijB[2, 2] - cijA[2, 2]) * x + cijA[2, 2])

        return a1, a2, a3, V, b1, b2, b3, qhklx, frac

    def RelaxationTriangle(self, hkl, sub, exp):
        """
        function which returns the relaxation triangle for a
        Alloy of given composition. Reciprocal space coordinates are
        calculated using the user-supplied experimental class

        Parameters
        ----------
        hkl : list or array-like
            Miller Indices
        sub : Crystal, or float
            substrate material or lattice constant
        exp : Experiment
            object from which the Transformation object and ndir are needed

        Returns
        -------
        qy, qz : float
            reciprocal space coordinates of the corners of the relaxation
            triangle

        """
        hkl = self._checkarray(hkl, "hkl")
        trans = exp._transform
        ndir = exp.ndir / VecNorm(exp.ndir)

        if isinstance(sub, Crystal):
            asub = sub.lattice.a
        elif isinstance(sub, float):
            asub = sub
        else:
            raise TypeError("Second argument (sub) must be of type float or "
                            "an instance of xrayutilities.materials.Crystal")

        # test if inplane direction of hkl is the same as the one for the
        # experiment otherwise warn the user
        hklinplane = VecCross(VecCross(exp.ndir, hkl), exp.ndir)
        if (VecNorm(VecCross(hklinplane, exp.idir)) > config.EPSILON):
            warnings.warn("Alloy: given hkl differs from the geometry of the "
                          "Experiment instance in the azimuthal direction")

        # transform elastic constants to correct coordinate frame
        cijA = Cijkl2Cij(trans(self.matA.cijkl, rank=4))
        cijB = Cijkl2Cij(trans(self.matB.cijkl, rank=4))

        a1, a2, a3, V, b1, b2, b3, qhklx, frac = self._definehelpers(hkl,
                                                                     cijA,
                                                                     cijB)

        qr_i = trans(qhklx(self.x))[1]
        qr_p = trans(qhklx(self.x))[2]
        qs_i = copysign(2*numpy.pi/asub * VecNorm(VecCross(ndir, hkl)), qr_i)
        qs_p = 2*numpy.pi/asub * abs(VecDot(ndir, hkl))

        # calculate pseudomorphic points for A and B
        def abulk(x):
            return math.VecNorm(a1(x))

        def aperp(x):
            return abulk(self.x) * (1 + frac(x) * (1 - asub / abulk(self.x)))

        qp_i = copysign(2*numpy.pi/asub * VecNorm(VecCross(ndir, hkl)), qr_i)
        qp_p = 2*numpy.pi/aperp(self.x) * abs(VecDot(ndir, hkl))

        # assembly return values
        qy = numpy.array([qr_i, qp_i, qs_i, qr_i], dtype=numpy.double)
        qz = numpy.array([qr_p, qp_p, qs_p, qr_p], dtype=numpy.double)

        return qy, qz


class CubicAlloy(Alloy):

    def __init__(self, matA, matB, x):
        # here one could check if material is really cubic!!
        Alloy.__init__(self, matA, matB, x)

    def ContentBsym(self, q_perp, hkl, inpr, asub, relax):
        """
        function that determines the content of B
        in the alloy from the reciprocal space position
        of a symetric peak. As an additional input the substrates
        lattice parameter and the degree of relaxation must be given

        Parameters
        ----------
        q_perp :    float
            perpendicular peak position of the reflection hkl of the alloy in
            reciprocal space
        hkl :       list
            Miller indices of the measured symmetric reflection (also defines
            the surface normal
        inpr :      list
            Miller indices of a Bragg peak defining the inplane reference
            direction
        asub :      float
            substrate lattice parameter
        relax :     float
            degree of relaxation (needed to obtain the content from symmetric
            reciprocal space position)

        Returns
        -------
        content :   float
            the content of B in the alloy determined from the input variables

        """

        # check input parameters
        q_perp = self._checkfinitenumber(q_perp, "q_perp")
        hkl = self._checkarray(hkl, "hkl")
        inpr = self._checkarray(inpr, "inpr")
        asub = self._checkfinitenumber(asub, "asub")
        relax = self._checkfinitenumber(relax, "relax")

        # calculate lattice constants from reciprocal space positions
        n = self.Q(hkl) / VecNorm(self.Q(hkl))
        # the following line is not generally true! only cubic materials
        aperp = 2 * numpy.pi / q_perp * abs(VecDot(n, hkl))

        # transform the elastic tensors to a coordinate frame attached to the
        # surface normal
        inp1 = VecCross(n, inpr) / VecNorm(VecCross(n, inpr))
        inp2 = VecCross(n, inp1)
        trans = math.CoordinateTransform(inp1, inp2, n)

        if config.VERBOSITY >= config.DEBUG:
            print("XU.materials.Alloy.ContentB: inp1/inp2: ", inp1, inp2)
        cijA = Cijkl2Cij(trans(self.matA.cijkl, rank=4))
        cijB = Cijkl2Cij(trans(self.matB.cijkl, rank=4))

        a1, a2, a3, V, b1, b2, b3, qhklx, frac = self._definehelpers(hkl,
                                                                     cijA,
                                                                     cijB)

        # the following line is not generally true! only cubic materials
        def abulk_perp(x):
            return abs(2 * numpy.pi / numpy.inner(qhklx(x), n) *
                       numpy.inner(n, hkl))

        # can we use abulk_perp here? for cubic materials this should work?!
        def ainp(x):
            return asub + relax * (abulk_perp(x) - asub)

        if config.VERBOSITY >= config.DEBUG:
            print("XU.materials.Alloy.ContentB: abulk_perp: %8.5g"
                  % (abulk_perp(0.)))

        def equation(x):
            return ((aperp - abulk_perp(x)) +
                    (ainp(x) - abulk_perp(x)) * frac(x))

        x = scipy.optimize.brentq(equation, -0.1, 1.1)

        return x

    def ContentBasym(self, q_inp, q_perp, hkl, sur):
        """
        function that determines the content of B
        in the alloy from the reciprocal space position
        of an asymmetric peak.

        Parameters
        ----------
        q_inp :     float
            inplane peak position of reflection hkl of the alloy in reciprocal
            space
        q_perp :    float
            perpendicular peak position of the reflection hkl of the alloy in
            reciprocal space
        hkl :       list
            Miller indices of the measured asymmetric reflection
        sur :       list
            Miller indices of the surface (determines the perpendicular
            direction)

        Returns
        -------
        content :   float
            content of B in the alloy determined from the input variables
        list
            [a_inplane a_perp, a_bulk_perp(x), eps_inplane, eps_perp];
            lattice parameters calculated from the reciprocal space positions
            as well as the strain (eps) of the layer
        """

        # check input parameters
        q_inp = self._checkfinitenumber(q_inp, "q_inp")
        q_perp = self._checkfinitenumber(q_perp, "q_perp")
        hkl = self._checkarray(hkl, "hkl")
        sur = self._checkarray(sur, "sur")

        # check if reflection is asymmetric
        if math.VecNorm(math.VecCross(self.Q(hkl), self.Q(sur))) < 1.e-8:
            raise InputError("Miller indices of a symmetric reflection were"
                             "given where an asymmetric reflection is needed")

        # calculate lattice constants from reciprocal space positions
        n = self.Q(sur) / VecNorm(self.Q(sur))
        q_hkl = self.Q(hkl)
        # the following two lines are not generally true! only cubic materials
        ainp = 2 * numpy.pi / abs(q_inp) * VecNorm(VecCross(n, hkl))
        aperp = 2 * numpy.pi / abs(q_perp) * abs(VecDot(n, hkl))

        # transform the elastic tensors to a coordinate frame attached to the
        # surface normal
        inp1 = VecCross(n, q_hkl) / VecNorm(VecCross(n, q_hkl))
        inp2 = VecCross(n, inp1)
        trans = math.CoordinateTransform(inp1, inp2, n)

        cijA = Cijkl2Cij(trans(self.matA.cijkl, rank=4))
        cijB = Cijkl2Cij(trans(self.matB.cijkl, rank=4))

        a1, a2, a3, V, b1, b2, b3, qhklx, frac = self._definehelpers(hkl,
                                                                     cijA,
                                                                     cijB)

        # the following two lines are not generally true! only cubic materials
        def abulk_inp(x):
            return abs(2 * numpy.pi / numpy.inner(qhklx(x), inp2) *
                       VecNorm(VecCross(n, hkl)))

        def abulk_perp(x):
            return abs(2 * numpy.pi / numpy.inner(qhklx(x), n) *
                       numpy.inner(n, hkl))

        if config.VERBOSITY >= config.DEBUG:
            print("XU.materials.Alloy.ContentB: abulk_inp/perp: %8.5g %8.5g"
                  % (abulk_inp(0.), abulk_perp(0.)))

        def equation(x):
            return ((aperp - abulk_perp(x)) +
                    (ainp - abulk_inp(x)) * frac(x))

        x = scipy.optimize.brentq(equation, -0.1, 1.1)

        eps_inplane = (ainp - abulk_perp(x)) / abulk_perp(x)
        eps_perp = (aperp - abulk_perp(x)) / abulk_perp(x)

        return x, [ainp, aperp, abulk_perp(x), eps_inplane, eps_perp]


def PseudomorphicMaterial(sub, layer, relaxation=0, trans=None):
    """
    This function returns a material whos lattice is pseudomorphic on a
    particular substrate material. The two materials must have similar unit
    cell definitions for the algorithm to work correctly, i.e. it does not work
    for combiniations of materials with different lattice symmetry. It is also
    crucial that the layer object includes values for the elastic tensor.

    Parameters
    ----------
    sub :       Crystal
        substrate material
    layer :     Crystal
        bulk material of the layer, including its elasticity tensor
    relaxation : float, optional
        degree of relaxation 0: pseudomorphic, 1: relaxed (default: 0)
    trans :     Tranform
        Transformation which transforms lattice directions into a surface
        orientated coordinate frame (x, y inplane, z out of plane). If None a
        (001) surface geometry of a cubic material is assumed.

    Returns
    -------
    An instance of Crystal holding the new pseudomorphically
    strained material.

    Raises
    ------
    InputError
        If the layer material has no elastic parameters
    """
    def get_inplane(lat):
        """determine inplane lattice parameter"""
        return (math.VecNorm(lat.GetPoint(trans.inverse((1, 0, 0)))) +
                math.VecNorm(lat.GetPoint(trans.inverse((0, 1, 0))))) / 2.

    if not trans:
        trans = math.Transform(numpy.identity(3))

    if numpy.all(layer.cijkl == 0):
        raise InputError("'layer' argument needs elastic parameters")

    # calculate the strain
    asub = get_inplane(sub.lattice)
    abulk = get_inplane(layer.lattice)
    apar = asub + (abulk - asub) * relaxation
    epar = (apar - abulk) / abulk
    cT = trans(layer.cijkl, rank=4)

    eperp = -epar * (cT[1, 1, 2, 2] + cT[2, 2, 0, 0]) / (cT[2, 2, 2, 2])
    eps = trans.inverse(numpy.diag((epar, epar, eperp)), rank=2)
    if config.VERBOSITY >= config.INFO_ALL:
        print("XU.materials.PseudomorphicMaterial: applying strain (inplane, "
              "perpendicular): %.4g %.4g" % (epar, eperp))

    # create the pseudomorphic material
    pmlatt = copy.deepcopy(layer.lattice)
    pmat = Crystal(layer.name, pmlatt, layer.cij)
    pmat.ApplyStrain(eps)
    return pmat



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014 Raphael Grifone <raphael.grifone@esrf.fr>
# Copyright (C) 2014-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
modules to help with the analysis of FastScan data acquired at the ESRF.
FastScan data are X-ray data (various detectors possible) acquired during
scanning the sample in real space with a Piezo Scanner.  The same functions
might be used to analze traditional SPEC mesh scans.

The module provides three core classes:

* FastScan
* FastScanCCD
* FastScanSeries

where the first two are able to parse single mesh/FastScans when one is
interested in data of a single channel detector or are detector and the last
one is able to parse full series of such mesh scans with either type of
detector

see examples/xrayutilities_kmap_ESRF.py for an example script
"""

import os.path
import re

import h5py
import numpy

from .. import config, utilities
from ..gridder import delta
from ..gridder2d import Gridder2D, Gridder2DList
from ..gridder3d import Gridder3D
from ..normalize import blockAverage2D
from .edf import EDFFile
from .spec import SPECFile


class FastScan(object):

    """
    class to help parsing and treating fast scan data.  FastScan is the
    aquisition of X-ray data while scanning the sample with piezo stages in
    real space. It's is available at several beamlines at the ESRF synchrotron
    light-source.
    """

    def __init__(self, filename, scannr,
                 xmotor='adcX', ymotor='adcY', path=""):
        """
        Constructor routine for the FastScan object. It initializes the object
        and parses the spec-scan for the needed data which are saved in
        properties of the FastScan object.

        Parameters
        ----------
        filename :      str
            file name of the fast scan spec file
        scannr :        int
            scannr of the to be parsed fast scan
        xmotor :        str, optional
            motor name of the x-motor (default: 'adcX' (ID01))
        ymotor :        str, optional
            motor name of the y-motor (default: 'adcY' (ID01))
        path :          str, optional
            optional path of the FastScan spec file
        """
        self.scannr = scannr
        self.xmotor = xmotor
        self.ymotor = ymotor

        if isinstance(filename, SPECFile):
            self.specfile = filename
            self.filename = self.specfile.filename
            self.full_filename = self.specfile.full_filename
            self.specscan = getattr(self.specfile, 'scan%d' % self.scannr)
        else:
            self.filename = filename
            self.full_filename = os.path.join(path, filename)
            self.filename = os.path.basename(self.full_filename)
            self.specscan = None

        # read the scan
        self.parse()

    def parse(self):
        """
        parse the specfile for the scan number specified in the constructor and
        store the needed informations in the object properties
        """

        # parse the file
        if not self.specscan:
            self.specfile = SPECFile(self.full_filename)
            self.specscan = getattr(self.specfile, 'scan%d' % self.scannr)
        self.specscan.ReadData()

        self.xvalues = self.specscan.data[self.xmotor]
        self.yvalues = self.specscan.data[self.ymotor]

        self.data = self.specscan.data

    def motorposition(self, motorname):
        """
        read the position of motor with name given by motorname from the data
        file. In case the motor is included in the data columns the returned
        object is an array with all the values from the file (although retrace
        clean is respected if already performed). In the case the motor is not
        moved during the scan only one value is returned.

        Parameters
        ----------
        motorname :     str
            name of the motor for which the position is wanted

        Returns
        -------
        ndarray
            motor position(s) of motor with name motorname during the scan
        """
        if self.specscan:
            # try reading value from data
            try:
                return self.data[motorname]
            except ValueError:
                try:
                    return self.specscan.init_motor_pos['INIT_MOPO_%s'
                                                        % motorname]
                except KeyError:
                    raise ValueError("given motorname '%s' not found in the "
                                     "Spec-data" % motorname)
        else:
            return None

    def retrace_clean(self):
        """
        function to clean the data of the scan from retrace artifacts created
        by the zig-zag scanning motion of the piezo actuators the function
        cleans the xvalues, yvalues and data attribute of the FastScan object.
        """

        # set window to determin the slope
        window = [-1, 0, 1]
        # calc the slope of x_motor movement using a window for better acuracy
        slope = numpy.convolve(self.xvalues, window, mode='same') / \
            numpy.convolve(numpy.arange(len(self.xvalues)), window, 'same')
        # select where slope is above the slope mean value
        # this can be modified if data points are missing of the retrace does
        # not clean all points
        mask = numpy.where(slope > slope.mean())

        # reduce data size by cutting out retrace
        self.xvalues = self.xvalues[mask]
        self.yvalues = self.yvalues[mask]
        self.data = self.data[mask]

    def grid2D(self, nx, ny, **kwargs):
        """
        function to grid the counter data and return the gridded X, Y and
        Intensity values.

        Parameters
        ----------
        nx, ny :    int
            number of bins in x, and y direction
        counter :   str, optional
            name of the counter to use for gridding (default: 'mpx4int' (ID01))
        gridrange : tuple, optional
            range for the gridder: format: ((xmin, xmax), (ymin, ymax))

        Returns
        -------
        Gridder2D
            Gridder2D object with X, Y, data on regular x, y-grid
        """
        self.counter = kwargs.get('counter', 'mpx4int')
        gridrange = kwargs.get('gridrange', None)

        # define gridder
        g2d = Gridder2D(nx, ny)
        if gridrange:
            g2d.dataRange(gridrange[0][0], gridrange[0][1],
                          gridrange[1][0], gridrange[1][1])

        # check if counter is in data fields
        if self.counter not in self.data.dtype.fields:
            raise ValueError("field named '%s' not found in data parsed from "
                             "scan #%d in file %s"
                             % (self.counter, self.scannr, self.filename))

        # grid data
        g2d(self.xvalues, self.yvalues, self.data[self.counter])

        # return gridded data
        return g2d


class FastScanCCD(FastScan):

    """
    class to help parsing and treating fast scan data including CCD frames.
    FastScan is the aquisition of X-ray data while scanning the sample with
    piezo stages in real space. It's is available at several beamlines at the
    ESRF synchrotron light-source. During such fast scan at every grid point
    CCD frames are recorded and need to be analyzed
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        imagefiletype : str, optional
            image file extension, either 'edf' / 'edf.gz' (default) or 'h5'

        other parameters are passed on to FastScanCCD
        """
        self.imagefiletype = kwargs.pop('imagefiletype', 'edf')
        self.imgfile = None
        self.nimages = None
        super().__init__(*args, **kwargs)

    def _getCCDnumbers(self, ccdnr):
        """
        internal function to return the ccd frame numbers from the data object
        or take them from the argument.
        """
        if isinstance(ccdnr, str):
            # check if counter is in data fields
            try:
                ccdnumbers = self.data[ccdnr]
            except ValueError:
                raise ValueError("field named '%s' not found in data parsed "
                                 "from scan #%d in file %s"
                                 % (ccdnr, self.scannr, self.filename))
        elif isinstance(ccdnr, (list, tuple, numpy.ndarray)):
            ccdnumbers = ccdnr
        else:
            raise ValueError("xu.FastScanCCD: wrong data type for "
                             "argument 'ccdnr'")
        return ccdnumbers

    def _gridCCDnumbers(self, nx, ny, ccdnr, gridrange=None):
        """
        internal function to grid the CCD frame number to produce a list of
        ccd-files per bin needed for the further treatment

        Parameters
        ----------
        nx, ny :        int
            number of bins in x, and y direction
        ccdnr :         str or array-like
            array with ccd file numbers of length length(FastScanCCD.data) OR a
            string with the data column name for the file ccd-numbers

        gridrange :     tuple, optional
            range for the gridder: format: ((xmin, xmax), (ymin, ymax))

        Returns
        -------
        gridder-object
            regular x, y-grid as well as 4-dimensional data object
        """
        g2l = Gridder2DList(nx, ny)
        if gridrange:
            g2l.dataRange(gridrange[0][0], gridrange[0][1],
                          gridrange[1][0], gridrange[1][1])

        ccdnumbers = self._getCCDnumbers(ccdnr)
        # assign ccd frames to grid
        g2l(self.xvalues, self.yvalues, ccdnumbers)

        return g2l

    def _read_image(self, filename, imgindex, nav, roi, filterfunc):
        """
        helper function to obtain one frame from an EDF/HDF5 file

        Parameters
        ----------
        filename :      str
            EDF file name
        imgindex :      int
            index of frame inside the given EDF file
        nav :           tuple or list
            number of detector pixel which will be averaged together (reduces
            the date size)
        roi :           tuple
            region of interest on the 2D detector. should be a list of lower
            and upper bounds of detector channels for the two pixel directions
            (default: None)
        filterfunc :    callable
            function applied to the CCD-frames before any processing. this
            function should take a single argument which is the ccddata which
            need to be returned with the same shape!  e.g. remove hot pixels,
            flat/darkfield correction

        Returns
        -------
        ndarray
            numpy 2D array with the detector frame
        """
        if roi is None:
            kwdict = {}
        else:
            kwdict = {'roi': roi}
        if 'edf' in self.imagefiletype:
            if not self.imgfile:
                self.imgfile = EDFFile(filename, keep_open=True)
            else:
                if self.imgfile.filename != filename:
                    self.imgfile = EDFFile(filename, keep_open=True)
            ccdfilt = self.imgfile.ReadData(imgindex)
        else:
            fileroot = os.path.splitext(os.path.splitext(filename)[0])[0]
            if not self.imgfile:
                self.imgfile = h5py.File(fileroot + '.h5', 'r')
            ccdfilt = self.imgfile.get(
                os.path.split(fileroot)[-1] + '_%04d' % imgindex).value
        if filterfunc:
            ccdfilt = filterfunc(ccdfilt)
        if roi is None and nav[0] == 1 and nav[1] == 1:
            return ccdfilt
        else:
            return blockAverage2D(ccdfilt, nav[0], nav[1], **kwdict)

    def _get_image_number(self, imgnum, imgoffset, fileoffset, ccdfiletmp):
        """
        function to obtain the image and file number. The logic for obtain this
        is likely to change between beamtimes.

        Parameters
        ----------
        imgnum :        int
            running image number from the data file
        imgoffset :     int
            offset in the image number
        fileoffset :    int
            offset in the file number
        ccdfiletmp :    str
            ccd file template string
        """
        if 'edf' in self.imagefiletype:
            if not self.imgfile:
                self.imgfile = EDFFile(ccdfiletmp % fileoffset, keep_open=True)
            if self.nimages is None:
                self.nimages = self.imgfile.nimages
        else:
            fileroot = os.path.splitext(os.path.splitext(ccdfiletmp
                                                         % fileoffset)[0])[0]
            if not self.imgfile:
                self.imgfile = h5py.File(fileroot + '.h5', 'r')
            if self.nimages is None:
                self.nimages = len(self.imgfile.items())
        filenumber = int((imgnum - imgoffset) // self.nimages + fileoffset)
        imgindex = int((imgnum - imgoffset) % self.nimages)
        return imgindex, filenumber

    def getccdFileTemplate(self, specscan, datadir=None, keepdir=0,
                           replacedir=None):
        """
        function to extract the CCD file template string from the comment
        in the SPEC-file scan-header.

        Parameters
        ----------
        specscan :      SpecScan
            spec-scan object from which header the CCD directory should be
            extracted
        datadir :       str, optional
            the CCD filenames are usually parsed from the scan object.  With
            this option the directory used for the data can be overwritten.
            Specify the datadir as simple string.  Alternatively the innermost
            directory structure can be automatically taken from the specfile.
            If this is needed specify the number of directories which should be
            kept using the keepdir option.
        keepdir :       int, optional
            number of directories which should be taken from the specscan.
            (default: 0)
        replacedir :    int, optional
            number of outer most directory names which should be replaced in
            the output (default = None). One can either give keepdir, or
            replacedir, with replace taking preference if both are given.

        Returns
        -------
        fmtstr :    str
            format string for the CCD file name using one number to build the
            real file name
        filenr :    int
            starting file number
        """
        hline = specscan.getheader_element('C imageFile')
        re_ccdfiles = re.compile(r'dir\[([a-zA-Z0-9_.%/]*)\] '
                                 r'prefix\[([a-zA-Z0-9_.%/]*)\] '
                                 r'idxFmt\[([a-zA-Z0-9_.%/]*)\] '
                                 r'nextNr\[([0-9]*)\] '
                                 r'suffix\[([a-zA-Z0-9_.%/]*)\]')
        m = re_ccdfiles.match(hline)
        if m:
            path, prefix, idxFmt, num, suffix = m.groups()
        else:
            ValueError('spec-scan does not contain images or the '
                       'corresponding header line is not detected correctly')
        ccdtmp = os.path.join(path, prefix + idxFmt + suffix)
        r = utilities.exchange_filepath(ccdtmp, datadir, keepdir, replacedir)
        return r, int(num)

    def getCCD(self, ccdnr, roi=None, datadir=None, keepdir=0,
               replacedir=None, nav=[1, 1], filterfunc=None):
        """
        function to read the ccd files and return the raw X, Y and DATA values.
        DATA represents a 3D object with first dimension representing the data
        point index and the remaining two dimensions representing detector
        channels

        Parameters
        ----------
        ccdnr :     array-like or str
            array with ccd file numbers of length length(FastScanCCD.data) OR a
            string with the data column name for the file ccd-numbers

        roi :       tuple, optional
            region of interest on the 2D detector. should be a list of lower
            and upper bounds of detector channels for the two pixel directions
            (default: None)
        datadir :   str, optional
            the CCD filenames are usually parsed from the SPEC file.  With this
            option the directory used for the data can be overwritten.  Specify
            the datadir as simple string.  Alternatively the innermost
            directory structure can be automatically taken from the specfile.
            If this is needed specify the number of directories which should be
            kept using the keepdir option.
        keepdir :   int, optional
            number of directories which should be taken from the SPEC file.
            (default: 0)
        replacedir : int, optional
            number of outer most directory names which should be replaced in
            the output (default = None). One can either give keepdir, or
            replacedir, with replace taking preference if both are given.
        nav :       tuple or list, optional
            number of detector pixel which will be averaged together (reduces
            the date size)
        filterfunc : callable
            function applied to the CCD-frames before any processing. this
            function should take a single argument which is the ccddata which
            need to be returned with the same shape!  e.g. remove hot pixels,
            flat/darkfield correction

        Returns
        -------
        X, Y :  ndarray
            x, y-array (1D)
        DATA :  ndarray
            3-dimensional data object
        """
        ccdnumbers = self._getCCDnumbers(ccdnr)

        ccdtemplate, nextNr = self.getccdFileTemplate(
            self.specscan, datadir, keepdir=keepdir, replacedir=replacedir)

        # read ccd shape from first image
        filename = ccdtemplate % nextNr
        ccdshape = self._read_image(filename, 0, nav, roi, filterfunc).shape
        ccddata = numpy.empty((self.xvalues.size, ccdshape[0], ccdshape[1]))
        if config.VERBOSITY >= config.INFO_ALL:
            print('XU.io.FastScanCCD: allocated ccddata array with %d bytes'
                  % ccddata.nbytes)

        # go through the ccd-frames
        for i, imgnum in enumerate(ccdnumbers):
            # read ccd-frames
            imgindex, filenumber = self._get_image_number(imgnum, nextNr,
                                                          nextNr, ccdtemplate)
            filename = ccdtemplate % filenumber
            ccd = self._read_image(filename, imgindex, nav, roi, filterfunc)
            ccddata[i, :, :] = ccd
        return self.xvalues, self.yvalues, ccddata

    def processCCD(self, ccdnr, roi, datadir=None, keepdir=0,
                   replacedir=None, filterfunc=None):
        """
        function to read a region of interest (ROI) from the ccd files and
        return the raw X, Y and intensity from ROI.

        Parameters
        ----------
        ccdnr :     array-like or str
            array with ccd file numbers of length length(FastScanCCD.data) OR a
            string with the data column name for the file ccd-numbers
        roi :       tuple or list
            region of interest on the 2D detector. Either a list of lower and
            upper bounds of detector channels for the two pixel directions as
            tuple or a list of mask arrays
        datadir :   str, optional
            the CCD filenames are usually parsed from the SPEC file.  With this
            option the directory used for the data can be overwritten.  Specify
            the datadir as simple string.  Alternatively the innermost
            directory structure can be automatically taken from the specfile.
            If this is needed specify the number of directories which should be
            kept using the keepdir option.
        keepdir :   int, optional
            number of directories which should be taken from the SPEC file.
            (default: 0)
        replacedir : int, optional
            number of outer most directory names which should be replaced in
            the output (default = None). One can either give keepdir, or
            replacedir, with replace taking preference if both are given.
        filterfunc : callable, optional
            function applied to the CCD-frames before any processing. this
            function should take a single argument which is the ccddata which
            need to be returned with the same shape!  e.g. remove hot pixels,
            flat/darkfield correction

        Returns
        -------
        X, Y, DATA :     ndarray
            x, y-array (1D) as well as 1-dimensional data object
        """
        ccdnumbers = self._getCCDnumbers(ccdnr)

        ccdtemplate, nextNr = self.getccdFileTemplate(
            self.specscan, datadir, keepdir=keepdir, replacedir=replacedir)

        if isinstance(roi, list):
            lmask = roi
            lroi = None
        else:
            lmask = [numpy.ones((roi[1]-roi[0], roi[3]-roi[2])), ]
            lroi = roi
        ccdroi = numpy.empty((len(lmask), self.xvalues.size))

        # go through the ccd-frames
        for i, imgnum in enumerate(ccdnumbers):
            # read ccd-frames
            imgindex, filenumber = self._get_image_number(imgnum, nextNr,
                                                          nextNr, ccdtemplate)
            filename = ccdtemplate % filenumber
            ccd = self._read_image(filename, imgindex, [1, 1], lroi,
                                   filterfunc)
            for j, m in enumerate(lmask):
                ccdroi[j, i] = numpy.sum(ccd[m])
        if len(lmask) == 1:
            return self.xvalues, self.yvalues, ccdroi[0]
        else:
            return self.xvalues, self.yvalues, ccdroi

    def gridCCD(self, nx, ny, ccdnr, roi=None, datadir=None, keepdir=0,
                replacedir=None, nav=[1, 1], gridrange=None, filterfunc=None):
        """
        function to grid the internal data and ccd files and return the gridded
        X, Y and DATA values. DATA represents a 4D object with first two
        dimensions representing X, Y and the remaining two dimensions
        representing detector channels

        Parameters
        ----------
        nx, ny :    int
            number of bins in x, and y direction
        ccdnr :     array-like or str
            array with ccd file numbers of length length(FastScanCCD.data) OR a
            string with the data column name for the file ccd-numbers

        roi :       tuple, optional
            region of interest on the 2D detector. should be a list of lower
            and upper bounds of detector channels for the two pixel directions
            (default: None)
        datadir :   str, optional
            the CCD filenames are usually parsed from the SPEC file.  With this
            option the directory used for the data can be overwritten.  Specify
            the datadir as simple string.  Alternatively the innermost
            directory structure can be automatically taken from the specfile.
            If this is needed specify the number of directories which should be
            kept using the keepdir option.
        keepdir :   int, optional
            number of directories which should be taken from the SPEC file.
            (default: 0)
        replacedir : int, optional
            number of outer most directory names which should be replaced in
            the output (default = None). One can either give keepdir, or
            replacedir, with replace taking preference if both are given.
        nav :       tuple or list, optional
            number of detector pixel which will be averaged together (reduces
            the date size)
        gridrange : tuple
            range for the gridder: format: ((xmin, xmax), (ymin, ymax))
        filterfunc : callable
            function applied to the CCD-frames before any processing. this
            function should take a single argument which is the ccddata which
            need to be returned with the same shape!  e.g. remove hot pixels,
            flat/darkfield correction

        Returns
        -------
        X, Y:       ndarray
            regular x, y-grid
        DATA :      ndarray
            4-dimensional data object
        """

        g2l = self._gridCCDnumbers(nx, ny, ccdnr, gridrange=gridrange)
        gdata = g2l.data

        ccdtemplate, nextNr = self.getccdFileTemplate(
            self.specscan, datadir, keepdir=keepdir, replacedir=replacedir)

        # read ccd shape from first image
        filename = ccdtemplate % nextNr
        ccdshape = self._read_image(filename, 0, nav, roi, filterfunc).shape
        ccddata = numpy.empty((self.xvalues.size, ccdshape[0], ccdshape[1]))
        if config.VERBOSITY >= config.INFO_ALL:
            print('XU.io.FastScanCCD: allocated ccddata array with %d bytes'
                  % ccddata.nbytes)

        # go through the gridded data and average the ccd-frames
        for i in range(gdata.shape[0]):
            for j in range(gdata.shape[1]):
                if not gdata[i, j]:
                    continue
                else:
                    framecount = 0
                    # read ccd-frames and average them
                    for imgnum in gdata[i, j]:
                        imgindex, filenumber = self._get_image_number(
                            imgnum, nextNr, nextNr, ccdtemplate)
                        filename = ccdtemplate % filenumber
                        ccd = self._read_image(filename, imgindex, nav,
                                               roi, filterfunc)
                        ccddata[i, j, ...] += ccd
                        framecount += 1
                    ccddata[i, j, ...] /= float(framecount)

        return g2l.xmatrix, g2l.ymatrix, ccddata


class FastScanSeries(object):

    """
    class to help parsing and treating a series of fast scan data including CCD
    frames.  FastScan is the aquisition of X-ray data while scanning the sample
    with piezo stages in real space. It's is available at several beamlines at
    the ESRF synchrotron light-source. During such fast scan at every grid
    point CCD frames are recorded and need to be analyzed.

    For the series of FastScans we assume that they are measured at different
    goniometer angles and therefore transform the data to reciprocal space.
    """

    def __init__(self, filenames, scannrs, nx, ny, *args, **kwargs):
        """
        Constructor routine for the FastScanSeries object. It initializes the
        object and creates a list of FastScanCCD objects. Importantly it also
        expects the motor names of the angles needed for reciprocal space
        conversion.

        Parameters
        ----------
        filenames : list or str
            file names of the fast scan spec files, in case of more than one
            filename supply a list of names and also a list of scan numbers for
            the different files in the 'scannrs' argument
        scannrs :   list
            scannrs of the to be parsed fast scans. in case of one specfile
            this is a list of numbers (e.g. [1, 2, 3]). when multiple filenames
            are given supply a separate list for every file (e.g.  [[1, 2,
            3],[2, 4]])
        nx, ny :    int
            grid-points for the real space grid
        args :      str
            motor names for the Reciprocal space conversion. The order needs be
            as required by the ``QConversion.area()`` function.
        xmotor :    str, optional
            motor name of the x-motor (default: 'adcX' (ID01))
        ymotor :    str, optional
            motor name of the y-motor (default: 'adcY' (ID01))
        ccdnr :     str, optional
            name of the ccd-number data column (default: 'imgnr' (ID01))
        counter :   str, optional
            name of a defined counter (roi) in the spec file (default:
            'mpx4int' (ID01))
        path :      str, optional
            path of the FastScan spec file (default: '')
        """

        if 'ccdnr' in kwargs:
            self.ccdnr = kwargs['ccdnr']
            kwargs.pop("ccdnr")
        else:
            self.ccdnr = 'imgnr'

        if 'counter' in kwargs:
            self.counter = kwargs['counter']
            kwargs.pop("counter")
        else:
            self.counter = 'mpx4int'

        if 'path' in kwargs:
            self.path = kwargs['path']
            kwargs.pop("path")
        else:
            self.path = ''

        self.fastscans = []
        self.nx = nx
        self.ny = ny
        self.motor_pos = None

        self.gonio_motors = []
        # save motor names
        for arg in args:
            if not isinstance(arg, str):
                raise ValueError("one of the motor name arguments is not of "
                                 "type 'str' but %s" % str(type(arg)))
            self.gonio_motors.append(arg)

        # create list of FastScans
        if isinstance(filenames, str):
            filenames = [filenames]
            scannrs = [scannrs]
        if isinstance(filenames, (tuple, list)):
            for fname in filenames:
                full_filename = os.path.join(self.path, fname)
                specfile = SPECFile(full_filename)
                for snrs in scannrs[filenames.index(fname)]:
                    self.fastscans.append(FastScanCCD(specfile,
                                                      snrs, **kwargs))
        else:
            raise ValueError("argument 'filenames' is not of "
                             "appropriate type!")

        self._init_minmax()
        for fs in self.fastscans:
            self._update_minmax(fs)

    def _init_minmax(self):
        self.gridded = False
        self.xmin = numpy.min(self.fastscans[0].xvalues)
        self.ymin = numpy.min(self.fastscans[0].yvalues)
        self.xmax = numpy.max(self.fastscans[0].xvalues)
        self.ymax = numpy.max(self.fastscans[0].yvalues)

    def _update_minmax(self, fs):
        if numpy.max(fs.xvalues) > self.xmax:
            self.xmax = numpy.max(fs.xvalues)
        if numpy.max(fs.yvalues) > self.ymax:
            self.ymax = numpy.max(fs.yvalues)
        if numpy.min(fs.xvalues) < self.xmin:
            self.xmin = numpy.min(fs.xvalues)
        if numpy.min(fs.yvalues) < self.ymin:
            self.ymin = numpy.min(fs.yvalues)

    def retrace_clean(self):
        """
        perform retrace clean for every FastScan in the series
        """
        self._init_minmax()

        for fs in self.fastscans:
            fs.retrace_clean()
            self._update_minmax(fs)

    def align(self, deltax, deltay):
        """
        Since a sample drift or shift due to rotation often occurs between
        different FastScans it should be corrected before combining them. Since
        determining such a shift is not straight-forward in general the user
        needs to supply the routine with the shifts in order correct the
        x, y-values for the different FastScans. Such a routine could for
        example use the integrated CCD intensities and determine the shift
        using a cross-convolution.

        Parameters
        ----------
        deltax, deltay :    list
            list of shifts in x/y-direction for every FastScan in the data
            structure
        """
        self._init_minmax()
        for fs in self.fastscans:
            i = self.fastscans.index(fs)
            fs.xvalues += deltax[i]
            fs.yvalues += deltay[i]
            self._update_minmax(fs)

    def read_motors(self):
        """
        read motor values from the series of fast scans
        """
        self.motor_pos = numpy.zeros((len(self.fastscans),
                                      len(self.gonio_motors)))
        for i in range(len(self.fastscans)):
            fs = self.fastscans[i]
            for j in range(len(self.gonio_motors)):
                mname = self.gonio_motors[j]
                self.motor_pos[i, j] = fs.motorposition(mname)

    def get_average_RSM(self, qnx, qny, qnz, qconv, datadir=None, keepdir=0,
                        replacedir=None, roi=None, nav=(1, 1),
                        filterfunc=None):
        """
        function to return the reciprocal space map data averaged over all x, y
        positions from a series of FastScan measurements. It necessary to give
        the QConversion-object to be used for the reciprocal space conversion.
        The QConversion-object is expected to have the 'area' conversion
        routines configured properly. This function needs to read all detector
        images, so be prepared to lean back for a moment!

        Parameters
        ----------
        qnx, qny, qnz : int
            number of points used for the 3D Gridder
        qconv :         QConversion
            QConversion-object to be used for the conversion of the CCD-data to
            reciprocal space

        roi :           tuple, optional
            region of interest on the 2D detector. should be a list of lower
            and upper bounds of detector channels for the two pixel directions
            (default: None)
        nav :           tuple or list, optional
            number of detector pixel which will be averaged together (reduces
            the date size)
        filterfunc :    callable, optional
            function applied to the CCD-frames before any processing. this
            function should take a single argument which is the ccddata which
            need to be returned with the same shape!  e.g. remove hot pixels,
            flat/darkfield correction
        datadir :       str, optional
            the CCD filenames are usually parsed from the SPEC file.  With this
            option the directory used for the data can be overwritten.  Specify
            the datadir as simple string.  Alternatively the innermost
            directory structure can be automatically taken from the specfile.
            If this is needed specify the number of directories which should be
            kept/replaced using the keepdir/replacedir option.
        keepdir :       int, optional
            number of directories which should be taken from the SPEC file.
            (default: 0)
        replacedir :    int, optional
            number of outer most directory names which should be replaced in
            the output (default = None). One can either give keepdir, or
            replacedir, with replace taking preference if both are given.

        Returns
        -------
        Gridder3D
            gridded reciprocal space map
        """
        if self.motor_pos is None:
            self.read_motors()

        # determine q-coordinates
        kwargs = {'Nav': nav}
        if roi:
            kwargs['roi'] = roi
        qx, qy, qz = qconv.area(*self.motor_pos.T, **kwargs)

        # define gridder with fixed optimized q-range
        g3d = Gridder3D(qnx, qny, qnz)
        g3d.keep_data = True
        g3d.dataRange(qx.min(), qx.max(), qy.min(),
                      qy.max(), qz.min(), qz.max(), fixed=True)

        # start parsing the images and grid the data frame by frame
        for fsidx, fsccd in enumerate(self.fastscans):
            ccdtemplate, nextNr = fsccd.getccdFileTemplate(
                fsccd.specscan, datadir, keepdir=keepdir,
                replacedir=replacedir)

            ccdnumbers = fsccd._getCCDnumbers(self.ccdnr)
            ccdav = numpy.zeros_like(qx[fsidx, ...])

            # go through the ccdframes
            for i, imgnum in enumerate(ccdnumbers):
                # read ccdframes
                imgindex, filenumber = fsccd._get_image_number(
                    imgnum, nextNr, nextNr, ccdtemplate)
                filename = ccdtemplate % filenumber
                ccd = fsccd._read_image(filename, imgindex, nav, roi,
                                        filterfunc)
                ccdav += ccd
            g3d(qx[fsidx, ...], qy[fsidx, ...], qz[fsidx, ...], ccdav)

        return g3d

    def get_sxrd_for_qrange(self, qrange, qconv, datadir=None, keepdir=0,
                            replacedir=None, roi=None, nav=(1, 1),
                            filterfunc=None):
        """
        function to return the real space data averaged over a certain q-range
        from a series of FastScan measurements. It necessary to give the
        QConversion-object to be used for the reciprocal space conversion.  The
        QConversion-object is expected to have the 'area' conversion routines
        configured properly.

        Note:
            This function assumes that all FastScans were performed in the same
            real space positions, no gridding or aligning is performed!

        Parameters
        ----------
        qrange :    list or tuple
            q-limits defining a box in reciprocal space.  six values are
            needed: [minx, maxx, miny, ..., maxz]
        qconv :     QConversion
            QConversion object to be used for the conversion of the CCD-data to
            reciprocal space

        roi :       tuple, optional
            region of interest on the 2D detector. should be a list of lower
            and upper bounds of detector channels for the two pixel directions
            (default: None)
        nav :       tuple or list, optional
            number of detector pixel which will be averaged together (reduces
            the date size)
        filterfunc : callable, optional
            function applied to the CCD-frames before any processing. this
            function should take a single argument which is the ccddata which
            need to be returned with the same shape!  e.g. remove hot pixels,
            flat/darkfield correction
        datadir :   str, optional
            the CCD filenames are usually parsed from the SPEC file.  With this
            option the directory used for the data can be overwritten.  Specify
            the datadir as simple string.  Alternatively the innermost
            directory structure can be automatically taken from the specfile.
            If this is needed specify the number of directories which should be
            kept/replaced using the keepdir/replacedir option.
        keepdir :   int, optional
            number of directories which should be taken from the SPEC file.
            (default: 0)
        replacedir : int, optional
            number of outer most directory names which should be replaced in
            the output (default = None). One can either give keepdir, or
            replacedir, with replace taking preference if both are given.

        Returns
        -------
        xvalues, yvalues, data :    ndarray
            x, y, and data values
        """
        if self.motor_pos is None:
            self.read_motors()

        # determine q-coordinates
        kwargs = {'Nav': nav}
        if roi:
            kwargs['roi'] = roi
        qx, qy, qz = qconv.area(*self.motor_pos.T, **kwargs)
        output = numpy.zeros_like(self.fastscans[0].xvalues)

        # parse the images only if some q coordinates fall into the ROI
        for fsidx, fsccd in enumerate(self.fastscans):
            mask = numpy.logical_and.reduce((
                qx[fsidx] > qrange[0], qx[fsidx] < qrange[1],
                qy[fsidx] > qrange[2], qy[fsidx] < qrange[3],
                qz[fsidx] > qrange[4], qz[fsidx] < qrange[5]))

            if numpy.any(mask):
                ccdtemplate, nextNr = fsccd.getccdFileTemplate(
                    fsccd.specscan, datadir, keepdir=keepdir,
                    replacedir=replacedir)

                ccdnumbers = fsccd._getCCDnumbers(self.ccdnr)

                # go through the ccdframes
                for i, imgnum in enumerate(ccdnumbers):
                    # read ccdframes
                    imgindex, filenumber = fsccd._get_image_number(
                        imgnum, nextNr, nextNr, ccdtemplate)
                    filename = ccdtemplate % filenumber
                    ccd = fsccd._read_image(filename, imgindex, nav, roi,
                                            filterfunc)
                    output[i] += numpy.sum(ccd[mask])

        return fsccd.xvalues, fsccd.yvalues, output

    def getCCDFrames(self, posx, posy, typ='real'):
        """
        function to determine the list of ccd-frame numbers for a specific real
        space position. The real space position must be within the data limits
        of the FastScanSeries otherwise an ValueError is thrown

        Parameters
        ----------
        posx :      float
            real space x-position or index in x direction
        posy :      float
            real space y-position or index in y direction

        typ :       {'real', 'index'}, optional
            type of coordinates. specifies if the position is specified as real
            space coordinate or as index. (default: 'real')

        Returns
        -------
        list
            ``[[motorpos1, ccdnrs1], [motorpos2, ccdnrs2], ...]``  where
            motorposN is from the N-ths FastScan in the series and ccdnrsN is
            the list of according CCD-frames
        """

        # determine grid point for position x, y
        if typ == 'real':
            # grid point calculation
            def gindex(x, min, delt):
                return numpy.round((x - min) / delt)

            xdelta = delta(self.xmin, self.xmax, self.nx)
            ydelta = delta(self.ymin, self.ymax, self.ny)
            xidx = gindex(posx, self.xmin, xdelta)
            yidx = gindex(posy, self.ymin, ydelta)
        elif typ == 'index':
            xidx = posx
            yidx = posy
        else:
            raise ValueError("given value of 'typ' is invalid.")

        if xidx >= self.nx or xidx < 0:
            raise ValueError("specified x-position is out of the data range")
        if yidx > self.ny or yidx < 0:
            raise ValueError("specified y-position is out of the data range")

        # read motor values and perform gridding for all subscans
        if not self.gridded:
            self.read_motors()
            self.glist = []
            for fs in self.fastscans:
                g2l = fs._gridCCDnumbers(
                    self.nx, self.ny, self.ccdnr,
                    gridrange=((self.xmin, self.xmax), (self.ymin, self.ymax)))
                self.glist.append(g2l)  # contains the ccdnumbers in g2l.data

            self.gridded = True

        # return the ccdnumbers and goniometer angles for this position
        ret = []
        for i in range(len(self.glist)):
            motorpos = self.motor_pos[i]
            ccdnrs = self.glist[i].data[xidx, yidx]
            ret.append([motorpos, ccdnrs])

        return ret

    def rawRSM(self, posx, posy, qconv, roi=None, nav=[1, 1], typ='real',
               datadir=None, keepdir=0, replacedir=None, filterfunc=None,
               **kwargs):
        """
        function to return the reciprocal space map data at a certain
        x, y-position from a series of FastScan measurements. It necessary to
        give the QConversion-object to be used for the reciprocal space
        conversion.  The QConversion-object is expected to have the 'area'
        conversion routines configured properly.

        Parameters
        ----------
        posx :      float
            real space x-position or index in x direction
        posy :      float
            real space y-position or index in y direction
        qconv :     QConversion
            QConversion-object to be used for the conversion of the CCD-data to
            reciprocal space

        roi :       tuple, optional
            region of interest on the 2D detector. should be a list of lower
            and upper bounds of detector channels for the two pixel directions
            (default: None)
        nav :       tuple or list, optional
            number of detector pixel which will be averaged together (reduces
            the date size)
        typ :       {'real', 'index'}, optional
            type of coordinates. specifies if the position is specified as real
            space coordinate or as index. (default: 'real')
        filterfunc : callable, optional
            function applied to the CCD-frames before any processing. this
            function should take a single argument which is the ccddata which
            need to be returned with the same shape!  e.g. remove hot pixels,
            flat/darkfield correction
        UB :        array-like, optional
            sample orientation matrix
        datadir :   str, optional
            the CCD filenames are usually parsed from the SPEC file.  With this
            option the directory used for the data can be overwritten.  Specify
            the datadir as simple string.  Alternatively the innermost
            directory structure can be automatically taken from the specfile.
            If this is needed specify the number of directories which should be
            kept using the keepdir option.
        keepdir :   int, optional
            number of directories which should be taken from the SPEC file.
            (default: 0)
        replacedir : int, optional
            number of outer most directory names which should be replaced in
            the output (default = None). One can either give keepdir, or
            replacedir, with replace taking preference if both are given.

        Returns
        -------
        qx, qy, qz :    ndarray
            reciprocal space positions of the reciprocal space map
        ccddata :       ndarray
            raw data of the reciprocal space map
        valuelist :     ndarray
            valuelist containing the ccdframe numbers and corresponding motor
            positions
        """
        U = kwargs.get('UB', numpy.identity(3))

        # get CCDframe numbers and motor values
        valuelist = self.getCCDFrames(posx, posy, typ)
        # load ccd frames and convert to reciprocal space
        fsccd = self.fastscans[0]
        ccdtemplate, nextNr = fsccd.getccdFileTemplate(
            fsccd.specscan, datadir, keepdir=keepdir, replacedir=replacedir)

        # read ccd shape from first image
        imgindex, filenumber = fsccd._get_image_number(
                        valuelist[0][1], nextNr, nextNr, ccdtemplate)
        filename = ccdtemplate % filenumber
        ccdshape = fsccd._read_image(filename, imgindex, nav, roi,
                                     filterfunc).shape
        ccddata = numpy.zeros((len(self.fastscans), ccdshape[0], ccdshape[1]))
        motors = []

        for i in range(len(self.gonio_motors)):
            motors.append(numpy.zeros(0))
        # go through the gridded data and average the ccdframes
        for i in range(len(self.fastscans)):
            imotors, ccdnrs = valuelist[i]
            fsccd = self.fastscans[i]
            # append motor positions
            for j in range(len(self.gonio_motors)):
                motors[j] = numpy.append(motors[j], imotors[j])
            # read CCD
            if not ccdnrs:
                continue
            else:
                ccdtemplate, nextNr = fsccd.getccdFileTemplate(fsccd.specscan)
                framecount = 0
                # read ccd-frames and average them
                for imgnum in ccdnrs:
                    imgindex, filenumber = fsccd._get_image_number(
                        imgnum, nextNr, nextNr, ccdtemplate)
                    filename = ccdtemplate % filenumber
                    ccd = fsccd._read_image(filename, imgindex, nav, roi,
                                            filterfunc)
                    ccddata[i, ...] += ccd
                    framecount += 1
                ccddata[i, ...] /= float(framecount)

        qx, qy, qz = qconv.area(*motors, roi=roi, Nav=nav, UB=U)
        return qx, qy, qz, ccddata, valuelist

    def gridRSM(self, posx, posy, qnx, qny, qnz, qconv, roi=None, nav=[1, 1],
                typ='real', filterfunc=None, **kwargs):
        """
        function to calculate the reciprocal space map at a certain
        x, y-position from a series of FastScan measurements it is necessary to
        specify the number of grid-oints for the reciprocal space map and the
        QConversion-object to be used for the reciprocal space conversion.  The
        QConversion-object is expected to have the 'area' conversion routines
        configured properly.

        Parameters
        ----------
        posx :      float
            real space x-position or index in x direction
        posy :      float
            real space y-position or index in y direction
        qnx, qny, qnz : int
            number of points in the Qx, Qy, Qz direction of the gridded
            reciprocal space map
        qconv :     QConversion
            QConversion-object to be used for the conversion of the CCD-data to
            reciprocal space
        roi :       tuple, optional
            region of interest on the 2D detector. should be a list of lower
            and upper bounds of detector channels for the two pixel directions
            (default: None)
        nav :       tuple or list, optional
            number of detector pixel which will be averaged together (reduces
            the date size)
        typ :       {'real', 'index'}, optional
            type of coordinates. specifies if the position is specified as real
            space coordinate or as index. (default: 'real')
        filterfunc : callable, optional
            function applied to the CCD-frames before any processing. this
            function should take a single argument which is the ccddata which
            need to be returned with the same shape!  e.g. remove hot pixels,
            flat/darkfield correction
        UB :         ndarray
            sample orientation matrix

        Returns
        -------
        Gridder3D
            object with gridded reciprocal space map
        """
        qx, qy, qz, ccddata, vallist = self.rawRSM(
            posx, posy, qconv, roi=roi, nav=nav,
            typ=typ, filterfunc=filterfunc, **kwargs)
        # perform 3D gridding and return the data or gridder
        g = Gridder3D(qnx, qny, qnz)
        g(qx, qy, qz, ccddata)
        return g

    def grid2Dall(self, nx, ny, **kwargs):
        """
        function to grid the counter data and return the gridded X, Y and
        Intensity values from all the FastScanSeries.

        Parameters
        ----------
        nx, ny :    int
            number of bins in x, and y direction

        counter :   str, optional
            name of the counter to use for gridding (default: 'mpx4int' (ID01))
        gridrange : tuple, optional
            range for the gridder: format: ((xmin, xmax), (ymin, ymax))

        Returns
        -------
        Gridder2D
            object with X, Y, data on regular x, y-grid
        """
        counter = kwargs.get('counter', 'mpx4int')
        gridrange = kwargs.get('gridrange', ((self.xmin, self.xmax),
                                             (self.ymin, self.ymax)))

        # define gridder
        g2d = Gridder2D(nx, ny)
        if gridrange:
            g2d.dataRange(gridrange[0][0], gridrange[0][1],
                          gridrange[1][0], gridrange[1][1])
        g2d.KeepData(True)

        for fs in self.fastscans:
            # check if counter is in data fields
            if counter not in fs.data.dtype.fields:
                raise ValueError("field named '%s' not found in data parsed "
                                 "from scan #%d in file %s"
                                 % (counter, fs.scannr, fs.filename))

            # grid data
            g2d(fs.xvalues, fs.yvalues, fs.data[counter])

        # return gridded data
        return g2d



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2017-2018 Dominik Kriegner <dominik.kriegner@gmail.com>
"""
module handling crystal lattice structures. A SGLattice consists of a space
group number and the position of atoms specified as Wyckoff positions along
with their parameters. Depending on the space group symmetry only certain
parameters of the resulting instance will be settable! A cubic lattice for
example allows only to set its 'a' lattice parameter but none of the other unit
cell shape parameters.
"""

import copy
import numbers
from collections import OrderedDict
from math import cos, radians, sin, sqrt

import numpy

from .. import math, utilities
from ..exception import InputError
from . import elements
from .atom import Atom
from .wyckpos import wp


class RangeDict(dict):
    def __getitem__(self, item):
        if type(item) != range:
            for key in self:
                if item in key:
                    return self[key]
        else:
            return super().__getitem__(item)


# space group number to symmetry and number of parameters dictionary
sgrp_sym = RangeDict({range(1, 3): ('triclinic', 6),
                      range(3, 16): ('monoclinic', 4),
                      range(16, 75): ('orthorhombic', 3),
                      range(75, 143): ('tetragonal', 2),
                      range(143, 168): ('trigonal', 2),
                      range(168, 195): ('hexagonal', 2),
                      range(195, 231): ('cubic', 1)})

sgrp_name = {'1': 'P1', '2': 'P-1', '3': 'P2', '4': 'P21', '5': 'C2',
             '6': 'Pm', '7': 'Pc', '8': 'Cm', '9': 'Cc', '10': 'P2/m',
             '11': 'P21/m', '12': 'C2/m', '13': 'P2/c', '14': 'P21/c',
             '15': 'C2/c', '16': 'P222', '17': 'P2221', '18': 'P21212',
             '19': 'P212121', '20': 'C2221', '21': 'C222', '22': 'F222',
             '23': 'I222', '24': 'I212121', '25': 'Pmm2', '26': 'Pmc21',
             '27': 'Pcc2', '28': 'Pma2', '29': 'Pca21', '30': 'Pnc2',
             '31': 'Pmn21', '32': 'Pba2', '33': 'Pna21', '34': 'Pnn2',
             '35': 'Cmm2', '36': 'Cmc21', '37': 'Ccc2', '38': 'Amm2',
             '39': 'Aem2', '40': 'Ama2', '41': 'Aea2', '42': 'Fmm2',
             '43': 'Fdd2', '44': 'Imm2', '45': 'Iba2', '46': 'Ima2',
             '47': 'Pmmm', '48': 'Pnnn', '49': 'Pccm', '50': 'Pban',
             '51': 'Pmma', '52': 'Pnna', '53': 'Pmna', '54': 'Pcca',
             '55': 'Pbam', '56': 'Pccn', '57': 'Pbcm', '58': 'Pnnm',
             '59': 'Pmmn', '60': 'Pbcn', '61': 'Pbca', '62': 'Pnma',
             '63': 'Cmcm', '64': 'Cmce', '65': 'Cmmm', '66': 'Cccm',
             '67': 'Cmme', '68': 'Ccce', '69': 'Fmmm', '70': 'Fddd',
             '71': 'Immm', '72': 'Ibam', '73': 'Ibca', '74': 'Imma',
             '75': 'P4', '76': 'P41', '77': 'P42', '78': 'P43',
             '79': 'I4', '80': 'I41', '81': 'P-4', '82': 'I-4',
             '83': 'P4/m', '84': 'P42/m', '85': 'P4/n', '86': 'P42/n',
             '87': 'I4/m', '88': 'I41/a', '89': 'P422', '90': 'P4212',
             '91': 'P4122', '92': 'P41212', '93': 'P4222', '94': 'P42212',
             '95': 'P4322', '96': 'P43212', '97': 'I422', '98': 'I4122',
             '99': 'P4mm', '100': 'P4bm', '101': 'P42cm', '102': 'P42nm',
             '103': 'P4cc', '104': 'P4nc', '105': 'P42mc', '106': 'P42bc',
             '107': 'I4mm', '108': 'I4cm', '109': 'I41md', '110': 'I41cd',
             '111': 'P-42m', '112': 'P-42c', '113': 'P-421m',
             '114': 'P-421c', '115': 'P-4m2', '116': 'P-4c2',
             '117': 'P-4b2', '118': 'P-4n2', '119': 'I-4m2',
             '120': 'I-4c2', '121': 'I-42m', '122': 'I-42d',
             '123': 'P4/mmm', '124': 'P4/mcc', '125': 'P4/nbm',
             '126': 'P4/nnc', '127': 'P4/mbm', '128': 'P4/mnc',
             '129': 'P4/nmm', '130': 'P4/ncc', '131': 'P42/mmc',
             '132': 'P42/mcm', '133': 'P42/nbc', '134': 'P42/nnm',
             '135': 'P42/mbc', '136': 'P42/mnm', '137': 'P42/nmc',
             '138': 'P42/ncm', '139': 'I4/mmm', '140': 'I4/mcm',
             '141': 'I41/amd', '142': 'I41/acd', '143': 'P3',
             '144': 'P31', '145': 'P32', '146': 'R3', '147': 'P-3',
             '148': 'R-3', '149': 'P312', '150': 'P321', '151': 'P3112',
             '152': 'P3121', '153': 'P3212', '154': 'P3221', '155': 'R32',
             '156': 'P3m1', '157': 'P31m', '158': 'P3c1', '159': 'P31c',
             '160': 'R3m', '161': 'R3c', '162': 'P-31m', '163': 'P-31c',
             '164': 'P-3m1', '165': 'P-3c1', '166': 'R-3m', '167': 'R-3c',
             '168': 'P6', '169': 'P61', '170': 'P65', '171': 'P62',
             '172': 'P64', '173': 'P63', '174': 'P-6', '175': 'P6/m',
             '176': 'P63/m', '177': 'P622', '178': 'P6122',
             '179': 'P6522', '180': 'P6222', '181': 'P6422',
             '182': 'P6322', '183': 'P6mm', '184': 'P6cc', '185': 'P63cm',
             '186': 'P63mc', '187': 'P-6m2', '188': 'P-6c2',
             '189': 'P-62m', '190': 'P-62c', '191': 'P6/mmm',
             '192': 'P6/mcc', '193': 'P63/mcm', '194': 'P63/mmc',
             '195': 'P23', '196': 'F23', '197': 'I23', '198': 'P213',
             '199': 'I213', '200': 'Pm-3', '201': 'Pn-3', '202': 'Fm-3',
             '203': 'Fd-3', '204': 'Im-3', '205': 'Pa-3', '206': 'Ia-3',
             '207': 'P432', '208': 'P4232', '209': 'F432',
             '210': 'F4132', '211': 'I432', '212': 'P4332',
             '213': 'P4132', '214': 'I4132', '215': 'P-43m',
             '216': 'F-43m', '217': 'I-43m', '218': 'P-43n',
             '219': 'F-43c', '220': 'I-43d', '221': 'Pm-3m',
             '222': 'Pn-3n', '223': 'Pm-3n', '224': 'Pn-3m',
             '225': 'Fm-3m', '226': 'Fm-3c', '227': 'Fd-3m',
             '228': 'Fd-3c', '229': 'Im-3m', '230': 'Ia-3d'}

sgrp_params = {'cubic:1': (('a', ), ('a', 'a', 'a', 90, 90, 90)),
               'cubic:2': (('a', ), ('a', 'a', 'a', 90, 90, 90)),
               'cubic': (('a', ), ('a', 'a', 'a', 90, 90, 90)),
               'hexagonal': (('a', 'c'), ('a', 'a', 'c', 90, 90, 120)),
               'trigonal:R': (('a', 'alpha'), ('a', 'a', 'a', 'alpha',
                                               'alpha', 'alpha')),
               'trigonal:H': (('a', 'c'), ('a', 'a', 'c', 90, 90, 120)),
               'trigonal': (('a', 'c'), ('a', 'a', 'c', 90, 90, 120)),
               'tetragonal:1': (('a', 'c'), ('a', 'a', 'c', 90, 90, 90)),
               'tetragonal:2': (('a', 'c'), ('a', 'a', 'c', 90, 90, 90)),
               'tetragonal': (('a', 'c'), ('a', 'a', 'c', 90, 90, 90)),
               'orthorhombic:1': (('a', 'b', 'c'),
                                  ('a', 'b', 'c', 90, 90, 90)),
               'orthorhombic:2': (('a', 'b', 'c'),
                                  ('a', 'b', 'c', 90, 90, 90)),
               'orthorhombic': (('a', 'b', 'c'),
                                ('a', 'b', 'c', 90, 90, 90)),
               'monoclinic:b': (('a', 'b', 'c', 'beta'),
                                ('a', 'b', 'c', 90, 'beta', 90)),
               'monoclinic:c': (('a', 'b', 'c', 'gamma'),
                                ('a', 'b', 'c', 90, 90, 'gamma')),
               'monoclinic': (('a', 'b', 'c', 'beta'),
                              ('a', 'b', 'c', 90, 'beta', 90)),
               'triclinic': (('a', 'b', 'c', 'alpha', 'beta', 'gamma'),
                             ('a', 'b', 'c', 'alpha', 'beta', 'gamma'))}


def get_possible_sgrp_suf(sgrp_nr):
    """
    determine possible space group suffix. Multiple suffixes might be possible
    for one space group due to different origin choice, unique axis, or choice
    of the unit cell shape.

    Parameters
    ----------
    sgrp_nr :   int
        space group number

    Returns
    -------
    str or list
        either an empty string or a list of possible valid suffix strings
    """
    sgrp_suf = ''
    if sgrp_nr in [3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]:
        sgrp_suf = [':b', ':c']
    elif sgrp_nr in [48, 50, 59, 68, 70, 85, 86, 88, 125, 126,
                     129, 130, 133, 134, 137, 138, 141, 142,
                     201, 203, 222, 224, 227, 228]:
        sgrp_suf = [':1', ':2']
    elif sgrp_nr in [146, 148, 155, 160, 161, 166, 167]:
        sgrp_suf = [':H', ':R']
    return sgrp_suf


def get_default_sgrp_suf(sgrp_nr):
    """
    determine default space group suffix
    """
    possibilities = get_possible_sgrp_suf(sgrp_nr)
    if possibilities:
        return possibilities[0]
    else:
        return ''


class WyckoffBase(list):

    """
    The WyckoffBase class implements a container for a set of Wyckoff positions
    that form the base of a crystal lattice. An instance of this class can be
    treated as a simple container object.
    """

    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)

    @staticmethod
    def _checkatom(atom):
        if isinstance(atom, str):
            atom = getattr(elements, atom)
        elif not isinstance(atom, Atom):
            raise TypeError("atom must be an instance of class "
                            "xrayutilities.materials.Atom")
        return atom

    @staticmethod
    def _checkpos(pos):
        if isinstance(pos, str):
            pos = (pos, None)
        elif isinstance(pos, (tuple, list)):
            if len(pos) == 1:
                pos = (pos[0], None)
            elif len(pos) == 2:
                if isinstance(pos[1], numbers.Number):
                    pos = (pos[0], (pos[1], ))
                elif pos[1] is None:
                    pos = (pos[0], None)
                else:
                    pos = (pos[0], tuple(pos[1]))
            else:
                if isinstance(pos[1], numbers.Number):
                    pos = (pos[0], tuple(pos[1:]))
        return pos

    def append(self, atom, pos, occ=1.0, b=0.):
        """
        add new Atom to the lattice base

        Parameters
        ----------
        atom :   Atom
            object to be added
        pos :    tuple or str
            Wyckoff position of the atom, along with its parameters.
            Examples: ('2i', (0.1, 0.2, 0.3)), or '1a'
        occ :    float, optional
            occupancy (default=1.0)
        b :      float, optional
            b-factor of the atom used as exp(-b*q**2/(4*pi)**2) to reduce the
            intensity of this atom (only used in case of temp=0 in
            StructureFactor and chi calculation)
        """
        atom = self._checkatom(atom)
        pos = self._checkpos(pos)
        list.append(self, (atom, pos, occ, b))

    def __setitem__(self, key, data):
        (atom, pos, occ, b) = data
        atom = self._checkatom(atom)
        pos = self._checkpos(pos)
        list.__setitem__(self, key, (atom, pos, float(occ), float(b)))

    def __copy__(self):
        """
        since we use a custom 'append' method we need to overwrite copy
        """
        cls = self.__class__
        new = cls.__new__(cls)
        for item in self:
            new.append(*item)
        return new

    def __deepcopy__(self, memo):
        cls = self.__class__
        new = cls.__new__(cls)
        memo[id(self)] = new
        for item in self:
            citem = copy.deepcopy(item, memo)
            new.append(*citem)
        return new

    def __str__(self):
        ostr = ''
        for i, (atom, p, occ, b) in enumerate(self):
            ostr += '%d: %s %s ' % (i, str(atom), p[0])
            if p[1] is not None:
                ostr += ' '.join(['%.4f' % v for v in p[1]])
            ostr += ' occ=%5.3f b=%5.3f\n' % (occ, b)
        return ostr

    def __contains__(self, item):
        """
        check if this list contains already the same element, at the same
        position, and with the same Debye Waller factor. The occupancy is not
        checked intentionally.

        Parameters
        ----------
        item :  tuple or list
            WyckoffBase entry to check if its present in this list

        Returns
        -------
        bool
        """
        for atom, p, occ, b in self:
            if (atom == item[0] and self.pos_eq(p, item[1]) and
                    numpy.isclose(b, item[3], atol=1e-4)):
                return True
        return False

    @staticmethod
    def entry_eq(e1, e2):
        """
        compare two entries including all its properties to be equal

        Parameters
        ----------
        e1, e2:     tuple
            tuples with length 4 containing the entries of WyckoffBase which
            should be compared
        """
        if (e1[0] == e2[0] and WyckoffBase.pos_eq(e1[1], e2[1]) and
                numpy.all(numpy.isclose(e1[2:], e2[2:], atol=1e-4))):
            return True
        return False

    @staticmethod
    def pos_eq(pos1, pos2):
        """
        compare Wyckoff positions

        Parameters
        ----------
        pos1, pos2:     tuple
            tuples with Wyckoff label and optional parameters
        """
        if pos1[0] != pos2[0]:
            return False
        if pos1[1] == pos2[1]:
            return True
        else:
            for f1, f2 in zip(pos1[1], pos2[1]):
                if not numpy.isclose(f1 % 1, f2 % 1, atol=1e-5):
                    return False
        return True

    def index(self, item):
        """
        return the index of the atom (same element, position, and Debye Waller
        factor). The occupancy is not checked intentionally. If the item is not
        present a ValueError is raised.

        Parameters
        ----------
        item :  tuple or list
            WyckoffBase entry

        Returns
        -------
        int
        """
        for i, (atom, p, occ, b) in enumerate(self):
            if (atom == item[0] and self.pos_eq(p, item[1]) and
                    numpy.isclose(b, item[3], atol=1e-4)):
                return i
        raise ValueError("%s is not in list" % str(item))


class SGLattice(object):
    """
    lattice object created from the space group number and corresponding unit
    cell parameters. atoms in the unit cell are specified by their Wyckoff
    position and their free parameters.

    this replaces the deprecated Lattice class
    """

    def __init__(self, sgrp, *args, **kwargs):
        """
        initialize class with space group number and atom list

        Parameters
        ----------
        sgrp :  int or str
            Space group number
        *args : float
            space group parameters. depending on the space group number this
            are 1 (cubic) to 6 (triclinic) parameters.
            cubic : a (lattice parameter).
            hexagonal : a, c.
            trigonal : a, c.
            tetragonal : a, c.
            orthorhombic : a, b, c.
            monoclinic : a, b, c, beta (in degree).
            triclinic : a, b, c, alpha, beta, gamma (in degree).
        atoms :     list, optional
            list of elements either as Element object or string with the
            element name. If you specify atoms you have to also give the same
            number of Wyckoff positions
        pos :       list, optional
            list of the atoms Wyckoff positions along with its parameters. If a
            position has no free parameter the parameters can be omitted.
            Example: [('2i', (0.1, 0.2, 0.3)), '1a']
        occ :       list, optional
            site occupation for the atoms. This is optional and defaults to 1
            if not given.
        b :         list, optional
            b-factor of the atom used as exp(-b*q**2/(4*pi)**2) to reduce the
            intensity of this atom (only used in case of temp=0 in
            StructureFactor and chi calculation)
        """
        valid_kwargs = {'atoms': 'list of elements',
                        'pos': 'list of Wyckoff positions',
                        'occ': 'site occupations',
                        'b': 'Debye Waller exponents'}
        utilities.check_kwargs(kwargs, valid_kwargs, self.__class__.__name__)
        self.space_group = str(sgrp)
        self.space_group_nr = int(self.space_group.split(':')[0])
        try:
            self.space_group_suf = ':' + self.space_group.split(':')[1]
        except IndexError:
            self.space_group_suf = get_default_sgrp_suf(self.space_group_nr)

        if self.space_group_suf != '':
            self.space_group = str(self.space_group_nr) + self.space_group_suf
        self.name = sgrp_name[str(self.space_group_nr)] + self.space_group_suf
        self.crystal_system, nargs = sgrp_sym[self.space_group_nr]
        self.crystal_system += self.space_group_suf
        if len(args) != nargs:
            raise ValueError('XU: number of parameters (%d) does not match the'
                             ' crystal symmetry (%s:%d)'
                             % (len(args), self.crystal_system, nargs))
        self.free_parameters = OrderedDict()
        for a, par in zip(args, sgrp_params[self.crystal_system][0]):
            self.free_parameters[par] = a

        self._parameters = OrderedDict()
        for i, p in enumerate(('a', 'b', 'c', 'alpha', 'beta', 'gamma')):
            key = sgrp_params[self.crystal_system][1][i]
            if isinstance(key, str):
                self._parameters[p] = self.free_parameters[key]
            else:
                self._parameters[p] = key

        # set atom positions in the lattice base
        self._wbase = WyckoffBase()
        atoms = kwargs.get('atoms', None)
        wps = kwargs.get('pos', None)
        if atoms:
            occs = kwargs.get('occ', [1.0, ] * len(atoms))
            bs = kwargs.get('b', [0.0, ] * len(atoms))
            for at, wpos, o, b in zip(atoms, wps, occs, bs):
                self._wbase.append(at, wpos, o, b)
        self.nsites = len(self._wbase)

        # define lattice vectors
        self._ai = numpy.zeros((3, 3))
        self._bi = numpy.empty((3, 3))
        a, b, c, alpha, beta, gamma = self._parameters.values()
        ra = radians(alpha)
        self._paramhelp = [cos(ra), cos(radians(beta)),
                           cos(radians(gamma)), sin(ra), 0]
        self._setlat()

    def base(self):
        """
        generator of atomic position within the unit cell.
        """
        if not self._wbase:
            return
        sgwp = wp[str(self.space_group)]
        for (atom, w, occ, b) in self._wbase:
            x, y, z = None, None, None
            parint, poslist = sgwp[w[0]]
            i = 0
            if parint & 1:
                try:
                    x = w[1][i]
                except TypeError:
                    print('XU.materials: Wyckoff position %s of %s needs '
                          'parameters (%d) -> wrong material definition'
                          % (w[0], str(self.space_group), parint))
                    raise
                i += 1
            if parint & 2:
                try:
                    y = w[1][i]
                except TypeError:
                    print('XU.materials: Wyckoff position %s of %s needs '
                          'parameters (%d) -> wrong material definition'
                          % (w[0], str(self.space_group), parint))
                    raise
                i += 1
            if parint & 4:
                try:
                    z = w[1][i]
                except TypeError:
                    print('XU.materials: Wyckoff position %s of %s needs '
                          'parameters (%d) -> wrong material definition'
                          % (w[0], str(self.space_group), parint))
                    raise
                i += 1
            if w[1]:
                if i != len(w[1]):
                    raise TypeError('XU.materials: too many parameters for '
                                    'Wyckoff position')

            for p in poslist:
                pos = eval(p, {'x': x, 'y': y, 'z': z})
                pos = numpy.asarray(pos)
                pos -= pos // 1
                yield atom, numpy.asarray(pos), occ, b

    def _setlat(self):
        a, b, c, alpha, beta, gamma = self._parameters.values()
        ca, cb, cg, sa, vh = self._paramhelp
        vh = sqrt(1 - ca**2-cb**2-cg**2 + 2*ca*cb*cg)
        self._paramhelp[4] = vh
        self._ai[0, 0] = a * vh / sa
        self._ai[0, 1] = a * (cg-cb*ca) / sa
        self._ai[0, 2] = a * cb
        self._ai[1, 1] = b * sa
        self._ai[1, 2] = b * ca
        self._ai[2, 2] = c
        self.transform = math.Transform(self._ai.T)
        self._setb()

    def _setb(self):
        V = self.UnitCellVolume()
        p = 2. * numpy.pi / V
        math.VecCross(p*self._ai[1, :], self._ai[2, :], out=self._bi[0, :])
        math.VecCross(p*self._ai[2, :], self._ai[0, :], out=self._bi[1, :])
        math.VecCross(p*self._ai[0, :], self._ai[1, :], out=self._bi[2, :])
        self.qtransform = math.Transform(self._bi.T)

    def _set_params_from_sym(self):
        for i, p in enumerate(('a', 'b', 'c', 'alpha', 'beta', 'gamma')):
            key = sgrp_params[self.crystal_system][1][i]
            if isinstance(key, str):
                if p not in self.free_parameters:
                    self._parameters[p] = self.free_parameters[key]

    @property
    def a(self):
        return self._parameters['a']

    @a.setter
    def a(self, value):
        if 'a' not in self.free_parameters:
            raise RuntimeError("a can not be set, its not a free parameter!")
        self._parameters['a'] = value
        self.free_parameters['a'] = value
        self._set_params_from_sym()
        self._setlat()

    @property
    def b(self):
        return self._parameters['b']

    @b.setter
    def b(self, value):
        if 'b' not in self.free_parameters:
            raise RuntimeError("b can not be set, its not a free parameter!")
        self._parameters['b'] = value
        self.free_parameters['b'] = value
        self._set_params_from_sym()
        self._setlat()

    @property
    def c(self):
        return self._parameters['c']

    @c.setter
    def c(self, value):
        if 'c' not in self.free_parameters:
            raise RuntimeError("c can not be set, its not a free parameter!")
        self._parameters['c'] = value
        self.free_parameters['c'] = value
        self._set_params_from_sym()
        self._setlat()

    @property
    def alpha(self):
        return self._parameters['alpha']

    @alpha.setter
    def alpha(self, value):
        if 'alpha' not in self.free_parameters:
            raise RuntimeError("alpha can not be set for this space group!")
        self._parameters['alpha'] = value
        self.free_parameters['alpha'] = value
        self._set_params_from_sym()
        ra = radians(value)
        self._paramhelp[0] = cos(ra)
        self._paramhelp[3] = sin(ra)
        self._setlat()

    @property
    def beta(self):
        return self._parameters['beta']

    @beta.setter
    def beta(self, value):
        if 'beta' not in self.free_parameters:
            raise RuntimeError("beta can not be set for this space group!")
        self._parameters['beta'] = value
        self.free_parameters['beta'] = value
        self._set_params_from_sym()
        self._paramhelp[1] = cos(radians(value))
        self._setlat()

    @property
    def gamma(self):
        return self._parameters['gamma']

    @gamma.setter
    def gamma(self, value):
        if 'gamma' not in self.free_parameters:
            raise RuntimeError("gamma can not be set for this space group!")
        self._parameters['gamma'] = value
        self.free_parameters['gamma'] = value
        self._set_params_from_sym()
        self._paramhelp[2] = cos(radians(value))
        self._setlat()

    def __eq__(self, other):
        """
        compare another SGLattice instance to decide if both are equal.
        To be equal they have to use the same space group, have equal lattice
        paramters and contain equal atoms in their base.
        """
        if self.space_group != other.space_group:
            return False
        # compare lattice parameters
        for prop in self.free_parameters:
            if getattr(self, prop) != getattr(other, prop):
                return False
        # compare atoms in base
        for e in self._wbase:
            if e not in other._wbase:
                return False
            idx = other._wbase.index(e)
            if not WyckoffBase.entry_eq(e, other._wbase[idx]):
                return False
        return True

    def GetPoint(self, *args):
        """
        determine lattice points with indices given in the argument

        Examples
        --------
        >>> xu.materials.Si.lattice.GetPoint(0, 0, 4)
        array([  0.     ,   0.     ,  21.72416])

        or

        >>> xu.materials.Si.lattice.GetPoint((1, 1, 1))
        array([ 5.43104,  5.43104,  5.43104])
        """
        if len(args) == 1:
            args = args[0]
        return self.transform(args)

    def GetQ(self, *args):
        """
        determine the reciprocal lattice points with indices given in the
        argument
        """
        if len(args) == 1:
            args = args[0]
        return self.qtransform(args)

    def GetHKL(self, *args):
        """
        determine the Miller indices of the given reciprocal lattice points
        """
        if len(args) == 1:
            args = args[0]
        return self.qtransform.inverse(args)

    def UnitCellVolume(self):
        """
        function to calculate the unit cell volume of a lattice (angstrom^3)
        """
        a, b, c, alpha, beta, gamma = self._parameters.values()
        return a * b * c * self._paramhelp[4]

    def ApplyStrain(self, eps):
        """
        Applies a certain strain on a lattice. The result is a change
        in the base vectors. The full strain matrix (3x3) needs to be given.

        Note:
            Here you specify the strain and not the stress -> NO elastic
            response of the material will be considered!

        Note:
            Although the symmetry of the crystal can be lowered by this
            operation the spacegroup remains unchanged! The 'free_parameters'
            attribute is, however, updated to mimic the possible reduction of
            the symmetry.

        Parameters
        ----------
        eps :    array-like
            a 3x3 matrix with all strain components
        """
        if isinstance(eps, (list, tuple)):
            eps = numpy.asarray(eps, dtype=numpy.double)
        if eps.shape != (3, 3):
            raise InputError("ApplyStrain needs a 3x3 matrix "
                             "with strain values")

        ai = self._ai + numpy.dot(eps, self._ai.T).T
        self._parameters['a'] = math.VecNorm(ai[0, :])
        self._parameters['b'] = math.VecNorm(ai[1, :])
        self._parameters['c'] = math.VecNorm(ai[2, :])
        self._parameters['alpha'] = math.VecAngle(ai[1, :], ai[2, :], deg=True)
        self._parameters['beta'] = math.VecAngle(ai[0, :], ai[2, :], deg=True)
        self._parameters['gamma'] = math.VecAngle(ai[0, :], ai[1, :], deg=True)
        # update helper parameters
        ra = radians(self._parameters['alpha'])
        self._paramhelp[0] = cos(ra)
        self._paramhelp[1] = cos(radians(self._parameters['beta']))
        self._paramhelp[2] = cos(radians(self._parameters['gamma']))
        self._paramhelp[3] = sin(ra)
        # set new transformations
        self._setlat()
        # update free_parameters
        for p, v in self.free_parameters.items():
            self.free_parameters[p] = self._parameters[p]
        # artificially reduce symmetry if needed
        for i, p in enumerate(('a', 'b', 'c', 'alpha', 'beta', 'gamma')):
            key = sgrp_params[self.crystal_system][1][i]
            if isinstance(key, str):
                if self._parameters[p] != self.free_parameters[key]:
                    self.free_parameters[p] = self._parameters[p]
            else:
                if self._parameters[p] != key:
                    self.free_parameters[p] = self._parameters[p]

    def isequivalent(self, hkl1, hkl2, equalq=False):
        """
        primitive way of determining if hkl1 and hkl2 are two
        crystallographical equivalent pairs of Miller indices

        Parameters
        ----------
        hkl1, hkl2 :    list
            Miller indices to be checked for equivalence
        equalq :        bool
            If False the length of the two q-vactors will be compared. If True
            it is assumed that the length of the q-vectors of hkl1 and hkl2 is
            equal!

        Returns
        -------
        bool
        """
        def leftShift(tup, n):
            try:
                n = n % len(tup)
            except ZeroDivisionError:
                return tuple()
            return tup[n:] + tup[0:n]

        def ispermutation(t1, t2):
            """returns true if t2 is an even permutation of t1"""
            if t1 == t2 or t1 == leftShift(t2, 1) or t1 == leftShift(t2, 2):
                return True
            else:
                return False

        def checkequal(hkl1, hkl2):
            if self.crystal_system.startswith('cubic'):
                if self.space_group_nr < 207:
                    if ispermutation(tuple(numpy.abs(hkl1)),
                                     tuple(numpy.abs(hkl2))):
                        return True
                    else:
                        return False
                else:
                    khl1 = (hkl1[1], hkl1[0], hkl1[2])
                    if (ispermutation(tuple(numpy.abs(hkl1)),
                                      tuple(numpy.abs(hkl2))) or
                        ispermutation(tuple(numpy.abs(khl1)),
                                      tuple(numpy.abs(hkl2)))):
                        return True
                    else:
                        return False
            elif self.crystal_system.startswith('hexagonal'):
                hki1 = (hkl1[0], hkl1[1], -hkl1[0] - hkl1[1])
                hki2 = (hkl2[0], hkl2[1], -hkl2[0] - hkl2[1])
                if (abs(hkl1[2]) == abs(hkl2[2]) and
                        sum(numpy.abs(hki1)) == sum(numpy.abs(hki2))):
                    return True
                else:
                    return False
            elif self.crystal_system.startswith('trigonal:R'):
                khl1 = (hkl1[1], hkl1[0], hkl1[2])
                if (ispermutation(hkl1, hkl2) or ispermutation(khl1, hkl2)):
                    return True
                else:
                    return False
            elif self.crystal_system.startswith('trigonal'):
                hki1 = (hkl1[0], hkl1[1], -hkl1[0] - hkl1[1])
                hki2 = (hkl2[0], hkl2[1], -hkl2[0] - hkl2[1])
                khi2 = (hkl2[1], hkl2[0], -hkl2[0] - hkl2[1])
                if ((hkl1[2] == hkl2[2] and ispermutation(hki1, hki2)) or
                        (hkl1[2] == -hkl2[2] and ispermutation(hki1, khi2))):
                    return True
                else:
                    return False
            elif self.crystal_system.startswith('tetragonal'):
                # this neglects that in some tetragonal materials hkl = khl
                hk1 = hkl1[:2]
                hk2 = hkl2[:2]
                if (abs(hkl1[2]) == abs(hkl2[2]) and
                        (hk1 == hk2 or hk1 == (-hk2[1], hk2[0]) or
                         hk1 == (-hk2[0], -hk2[1]) or
                         hk1 == (hk2[1], -hk2[0]))):
                    return True
                else:
                    return False
            elif self.crystal_system.startswith('orthorhombic'):
                if numpy.all(numpy.abs(hkl1) == numpy.abs(hkl2)):
                    return True
                else:
                    return False
            elif self.crystal_system.startswith('monoclinic:c'):
                hk1 = (hkl1[0], hkl1[1])
                hk2 = (hkl2[0], hkl2[1])
                if (abs(hkl1[2]) == abs(hkl2[2]) and
                        (hk1 == hk2 or hk1 == (-hk2[0], -hk2[1]))):
                    return True
                else:
                    return False
            elif self.crystal_system.startswith('monoclinic'):
                hl1 = (hkl1[0], hkl1[2])
                hl2 = (hkl2[0], hkl2[2])
                if (abs(hkl1[1]) == abs(hkl2[1]) and
                        (hl1 == hl2 or hl1 == (-hl2[0], -hl2[1]))):
                    return True
                else:
                    return False
            elif self.crystal_system.startswith('triclinic'):
                if (hkl1[0] == -hkl2[0] and hkl1[1] == -hkl2[1] and
                        hkl1[2] == -hkl2[2]):
                    return True
                else:
                    return False

        if not equalq:
            if not numpy.isclose(math.VecNorm(self.GetQ(hkl1)),
                                 math.VecNorm(self.GetQ(hkl2))):
                return False
        return checkequal(tuple(hkl1), tuple(hkl2))

    def __str__(self):
        ostr = "{sg} {cs} {n}: a = {a:.4f}, b = {b:.4f} c= {c:.4f}\n" +\
               "alpha = {alpha:.3f}, beta = {beta:.3f}, gamma = {gamma:.3f}\n"
        ostr = ostr.format(sg=self.space_group, cs=self.crystal_system,
                           n=self.name, **self._parameters)
        if self._wbase:
            ostr += "Lattice base:\n"
            ostr += str(self._wbase)
        return ostr

    @classmethod
    def convert_to_P1(cls, sglat):
        """
        create a P1 equivalent of the given SGLattice instance.

        Parameters
        ----------
        sglat :   SGLattice
            space group lattice instance to be converted to P1.

        Returns
        -------
        SGLattice
            instance with the same properties as sglat, however in the P1
            setting.
        """
        a, b, c, alpha, beta, gamma = (sglat.a, sglat.b, sglat.c, sglat.alpha,
                                       sglat.beta, sglat.gamma)
        atoms = []
        pos = []
        occ = []
        biso = []
        for at, p, o, bf in sglat.base():
            atoms.append(at)
            pos.append(('1a', p))
            occ.append(o)
            biso.append(bf)
        return cls(1, a, b, c, alpha, beta, gamma, atoms=atoms, pos=pos,
                   occ=occ, b=biso)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2017 Dominik Kriegner <dominik.kriegner@gmail.com>
"""
Defines new matplotlib Sqrt scale which further allows for negative values by
using the sign of the original value as sign of the plotted value.
"""

import math

import matplotlib
import numpy
from matplotlib import scale as mscale
from matplotlib import ticker as mticker
from matplotlib import transforms as mtransforms
from pkg_resources import parse_version


class SqrtAllowNegScale(mscale.ScaleBase):
    """
    Scales data using a sqrt-function, however, allowing also negative values.

    The scale function:
      sign(y) * sqrt(abs(y))

    The inverse scale function:
      sign(y) * y**2
    """
    name = 'sqrt'

    def __init__(self, axis, **kwargs):
        """
        Any keyword arguments passed to ``set_xscale`` and
        ``set_yscale`` will be passed along to the scale's
        constructor.
        """
        if parse_version(matplotlib.__version__) < parse_version('3.1.0'):
            super().__init__(**kwargs)
        else:
            super().__init__(axis, **kwargs)

    def set_default_locators_and_formatters(self, axis):
        axis.set_major_locator(SqrtTickLocator())

    def limit_range_for_scale(self, vmin, vmax, minpos):
        """
        Override to limit the bounds of the axis to the domain of the
        transform.  In the case of Mercator, the bounds should be
        limited to the threshold that was passed in.  Unlike the
        autoscaling provided by the tick locators, this range limiting
        will always be adhered to, whether the axis range is set
        manually, determined automatically or changed through panning
        and zooming.
        """
        return vmin, vmax

    def get_transform(self):
        return self.SqrtTransform()

    class SqrtTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True

        def transform_non_affine(self, a):
            """
            This transform takes an Nx1 ``numpy`` array and returns a
            transformed copy.
            """
            return numpy.sign(a) * numpy.sqrt(numpy.abs(a))

        def inverted(self):
            """
            return the inverse transform for this transform.
            """
            return SqrtAllowNegScale.InvertedSqrtTransform()

    class InvertedSqrtTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True

        def transform_non_affine(self, a):
            return numpy.sign(a) * a**2

        def inverted(self):
            return SqrtAllowNegScale.SqrtTransform()


class SqrtTickLocator(mticker.Locator):
    def __init__(self, nbins=7, symmetric=True):
        if parse_version(matplotlib.__version__) < parse_version('3.1.0'):
            self._base = mticker.Base(1.0)
        else:
            self._base = mticker._Edge_integer(1.0, 0)
        self.set_params(nbins, symmetric)

    def set_params(self, nbins, symmetric):
        """Set parameters within this locator."""
        self._nbins = nbins
        self._symmetric = symmetric

    def __call__(self):
        'Return the locations of the ticks'
        vmin, vmax = self.axis.get_view_interval()
        return self.tick_values(vmin, vmax)

    def tick_values(self, vmin, vmax):
        if vmax < vmin:
            vmin, vmax = vmax, vmin
        tmin = math.copysign(math.sqrt(abs(vmin)), vmin)
        tmax = math.copysign(math.sqrt(abs(vmax)), vmax)
        delta = (tmax - tmin) / self._nbins
        locs = numpy.arange(tmin, tmax, self._base.ge(delta))
        if self._symmetric and numpy.sign(tmin) != numpy.sign(tmax):
            locs -= locs[numpy.argmin(numpy.abs(locs))]
        locs = numpy.sign(locs) * locs**2
        return self.raise_if_exceeds(locs)

    def view_limits(self, dmin, dmax):
        """
        Set the view limits to the nearest multiples of base that
        contain the data
        """
        vmin = self._base.le(math.copysign(math.sqrt(abs(dmin)), dmin))
        vmax = self._base.ge(math.sqrt(dmax))
        if vmin == vmax:
            vmin -= 1
            vmax += 1

        return mtransforms.nonsingular(math.copysign(vmin**2, vmin), vmax**2)


# register new scale to matplotlib
mscale.register_scale(SqrtAllowNegScale)
